import sys
import os
import logging
import subprocess
import textwrap
from warnings import warn
from enum import Enum
from collections import Counter
import numpy as np
import itertools
from scipy.spatial import Voronoi
import h5py
import xmltodict
import pyvista as pv
import dopyqo
from dopyqo import plotting
from dopyqo.colors import *


class Wfc:
    """Class storing wavefunction and crystal information from Quantum ESPRESSO output files

    See from_file(...) function to create object from Quantum ESPRESSO output files.
    """

    def __init__(
        self,
        ik: int,
        xk: np.ndarray,
        ispin: int,
        gamma_only: bool,
        scalef: float,
        ngw: int,
        igwx: int,
        npol: int,
        nbnd: int,
        b1: np.ndarray,
        b2: np.ndarray,
        b3: np.ndarray,
        mill: np.ndarray,
        evc: np.ndarray,
        output_xml: str,
        pseudopots: list[dopyqo.Pseudopot] = None,
        kpoint_idx: int | None = None,
    ):
        """Class storing wavefunction and crystal information from Quantum ESPRESSO output files

        Args:
            ik (int): k-point index starting from 1
            xk (np.ndarray): k-point coordinates
            ispin (int): Spin index for LSDA case: ispin=1 for spin-up, ispin=2 for spin-down.
                          For unpolarized or non-colinear cases, ispin=1
            gamma_only (bool): True if a gamma-only calculation was performed, then only half of the plane waves are written to file
            scalef (float): Scale factor applied to wavefunctions
            ngw (int): Number of plane waves
            igwx (int): Max number of plane waves
            npol (int): Number of spin states for plane waves: 2 for non-colinear case, 1 otherwise
            nbnd (int): Number of Kohn-Sham orbitals, i.e., number of bands
            b1 (np.ndarray): First reciprocal lattice vector
            b2 (np.ndarray): Second reciprocal lattice vector
            b3 (np.ndarray): Third reciprocal lattice vector
            mill (np.ndarray): Miller indices for each plane wave. Shape (# of plane waves, 3)
            evc (np.ndarray): Coefficients describing the Kohn-Sham orbitals in terms of plane waves.
                              Shape (# of plane waves, # of Kohn-Sham orbitals)
            output_xml (str): Path to the XML-file Quantum ESPRESSO outputs
            pseudopots (list[Pseudopot], optional): List of dopyqo.Pseudopot objects, one object per atom type. Defaults to None.
            kpoint_idx (int | None, optional): k-point index starting from 0. Is set to 0, if None and Quantum ESPRESSO calculation
                                               only involved one k-point. Has to match ik+1, if Quantum ESPRESSO calculation
                                               involved multiple k-points. Defaults to None.
        """
        # NOTE: For gamma-only calculations the KS-orbitals are real in real-space.
        #       For k-point calculations, the gamma-point KS-orbitals are NOT real in real-space.
        #       This is because the KS-orbitals at the gamma-point do not have to be real in real-space,
        #       they can just be chosen to be real!
        self.ik = ik
        # k-point vector is in units of 1/bohr = 1/(5.29177210544e-11 meter)
        self.kpoint = xk
        self.ispin = ispin
        self.gamma_only = gamma_only
        self.scalef = scalef
        if not np.isclose(self.scalef, 1.0):
            print(f"{ORANGE}Warning: Scale-factor is not 1.0. Please contact a developer!{RESET_COLOR}")
        self.ngw = ngw
        self.igwx = igwx
        self.npol = npol
        self.nbnd = nbnd
        # Reciprocal lattice vectors are in units of 1/bohr = 1/(5.29177210544e-11 meter)
        self.b1 = b1
        self.b2 = b2
        self.b3 = b3
        self.mill = mill
        self.evc = evc
        self.pseudopots = pseudopots
        self.kpoint_idx = kpoint_idx

        if self.kpoint_idx is not None:
            assert isinstance(
                self.kpoint_idx, int
            ), f"kpoint_idx invalid! Expected integer or None but got: {self.kpoint_idx} of type {type(self.kpoint_idx)}"

        self.b = np.array([self.b1, self.b2, self.b3])  # b1 in first row, b2 in second row, b3 in third row.

        self.G = np.einsum("ij,jk->ik", mill, self.b)

        self.k_plus_G = xk + self.G
        self.k_plus_G_norm = np.linalg.norm(self.k_plus_G, ord=2, axis=1)

        # If only Γ point is sampled, only positive half of
        # the plane wave expansion coefficients are saved. Generate and append negative half here.
        # See https://docs.abinit.org/theory/wavefunctions/#plane wave-basis-set-sphere
        if self.gamma_only:
            self.evc_org = self.evc.copy()
            self.G_org = self.G.copy()
            self.mill_org = self.mill.copy()

            assert (self.G[0] == np.array([0.0, 0.0, 0.0])).all(), (
                f"Expected first G-vector to be the zero-vector but found {self.G[0]} in order" " to generate coefficients of negative G-vectors!"
            )
            self.G = np.append(self.G, -self.G[1:], axis=0)
            self.mill = np.append(self.mill, -self.mill[1:], axis=0)
            self.k_plus_G = xk + self.G
            self.k_plus_G_norm = np.linalg.norm(self.k_plus_G, ord=2, axis=1)
            self.evc = np.append(
                self.evc, self.evc[:, 1:].conj(), axis=1
            )  # conj() correct, see https://docs.abinit.org/theory/wavefunctions/#plane wave-basis-set-sphere

        with open(output_xml, encoding="utf-8") as file:
            self.xml_dict = xmltodict.parse(file.read())

        # | k + G | < G_cut with E_cut = G_cut^2/2 in atomic units (-> G_cut = \sqrt(2*E_cut))
        # see: https://wiki.fysik.dtu.dk/gpaw/tutorialsexercises/structureoptimization/stress/stress.html
        self.ecutwfc = float(self.xml_dict["qes:espresso"]["input"]["basis"]["ecutwfc"])
        self.gcutwfc = np.sqrt(self.ecutwfc * 2)
        self.ecutrho = float(self.xml_dict["qes:espresso"]["input"]["basis"]["ecutrho"])
        self.gcutrho = np.sqrt(self.ecutrho * 2)
        self.occupations_scf = self.xml_dict["qes:espresso"]["input"]["bands"]["occupations"]
        self.degauss = None
        self.smearing = None
        if self.occupations_scf == "smearing":
            smearing = self.xml_dict["qes:espresso"]["input"]["bands"]["smearing"]
            self.degauss = float(smearing["@degauss"])
            self.smearing = smearing["#text"]
        self.conv_thr = float(self.xml_dict["qes:espresso"]["input"]["electron_control"]["conv_thr"])
        self.mixing_beta = float(self.xml_dict["qes:espresso"]["input"]["electron_control"]["mixing_beta"])
        self.electron_maxstep = int(self.xml_dict["qes:espresso"]["input"]["electron_control"]["max_nstep"])

        # Read k-points
        self.n_kpoints = int(self.xml_dict["qes:espresso"]["output"]["band_structure"]["nks"])
        # If multiple k-points are available, you have to specify kpoint_idx
        if self.n_kpoints != 1 and self.kpoint_idx is None:
            print(
                f"{RED}Wavefunction error: You try loading a DFT calculation involving multiple k-points ({self.n_kpoints}) "
                + "without specifying the k-point you want to load (kpoint_idx=None). "
                + "Specify a k-point by setting kpoint_idx to the corresponding index of the k-point, starting at zero. "
                + "You can also directly load all k-points with the "
                + f"{self.__class__.__name__}.from_file_all_kpoints method.{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        # Set kpoint_idx to zero if only one k-point is available and kpoint_idx is None
        if self.n_kpoints == 1 and self.kpoint_idx is None:
            self.kpoint_idx = 0
        if self.kpoint_idx >= self.n_kpoints or self.kpoint_idx < 0 or not np.isclose(self.kpoint_idx, int(self.kpoint_idx)):
            raise ValueError(
                f"kpoint_idx ({self.kpoint_idx}) is invalid. It has to be an integer greater than "
                + f"zero and smaller than the number of k-points ({self.n_kpoints})."
            )
        assert self.kpoint_idx == self.ik - 1, (
            f"kpoint_idx ({self.kpoint_idx}) does not match the k-point index in the wavefunction "
            + f"({self.ik-1}). The wavefunction does not belong to the specified k-point index kpoint_idx ({self.kpoint_idx})."
        )

        # If calculation had more than one k-point xml_dict[...]["ks_energies"] is a list instead of a dict
        # self.kpoint_xml is in units of 2*np.pi/alat_out, NOT in 1/bohr = 1/(5.29177210544e-11 meter)
        if self.n_kpoints == 1:
            self.kpoint_xml = np.fromstring(
                self.xml_dict["qes:espresso"]["output"]["band_structure"]["ks_energies"]["k_point"]["#text"],
                sep=" ",
                dtype=np.float64,
            )
            self.kpoint_weight = float(self.xml_dict["qes:espresso"]["output"]["band_structure"]["ks_energies"]["k_point"]["@weight"])
        else:
            self.kpoint_xml = np.fromstring(
                self.xml_dict["qes:espresso"]["output"]["band_structure"]["ks_energies"][self.kpoint_idx]["k_point"]["#text"],
                sep=" ",
                dtype=np.float64,
            )
            self.kpoint_weight = float(
                self.xml_dict["qes:espresso"]["output"]["band_structure"]["ks_energies"][self.kpoint_idx]["k_point"]["@weight"]
            )

        fft_grid_x = int(self.xml_dict["qes:espresso"]["output"]["basis_set"]["fft_grid"]["@nr1"])
        fft_grid_y = int(self.xml_dict["qes:espresso"]["output"]["basis_set"]["fft_grid"]["@nr2"])
        fft_grid_z = int(self.xml_dict["qes:espresso"]["output"]["basis_set"]["fft_grid"]["@nr3"])
        self.fft_grid = np.array([fft_grid_x, fft_grid_y, fft_grid_z])

        self.spin = 1
        if (
            "nbnd_up" in self.xml_dict["qes:espresso"]["output"]["band_structure"]
            and "nbnd_dw" in self.xml_dict["qes:espresso"]["output"]["band_structure"]
        ):
            self.spin = 2

        self.etot = float(self.xml_dict["qes:espresso"]["output"]["total_energy"]["etot"])
        self.eband = float(self.xml_dict["qes:espresso"]["output"]["total_energy"]["eband"])
        self.ehart = float(self.xml_dict["qes:espresso"]["output"]["total_energy"]["ehart"])
        self.vtxc = float(self.xml_dict["qes:espresso"]["output"]["total_energy"]["vtxc"])
        self.etxc = float(self.xml_dict["qes:espresso"]["output"]["total_energy"]["etxc"])
        self.ewald = float(self.xml_dict["qes:espresso"]["output"]["total_energy"]["ewald"])
        atoms_dict = self.xml_dict["qes:espresso"]["input"]["atomic_structure"]["atomic_positions"]["atom"]

        self.atomic_species = self.xml_dict["qes:espresso"]["input"]["atomic_species"]["species"]
        if not isinstance(self.atomic_species, list):
            self.atomic_species = [self.atomic_species]
        self.atomic_species = {x["@name"]: {"mass": float(x["mass"]), "pseudo_file": x["pseudo_file"]} for x in self.atomic_species}

        # If the lattice vectors are defined in units of alat (CELL_PARAMETERS alat) the "input" and "output"
        # values of alat are different and represent alat (\sqrt(a1 @ a1)) calculated with the lattice vectors
        # in hartree/bohr units and alat defined in the input file (whyever this should be written in the "output" section)
        # via celldem(1) or A, respectively
        # The k-points in the wfc files are in units of the "output"-alat, but the alat of the unit cell is the
        # "input"-alat. Very confusing!
        self.alat = float(self.xml_dict["qes:espresso"]["input"]["atomic_structure"]["@alat"])
        alat_out = float(self.xml_dict["qes:espresso"]["output"]["atomic_structure"]["@alat"])
        assert np.allclose(self.kpoint_xml, self.kpoint / (2 * np.pi) * alat_out), (
            f"The k-point with index {self.kpoint_idx} from the xml-file ({self.kpoint_xml}) does not match the "
            + f"k-point in the wavefunction file ({self.kpoint / (2*np.pi) * alat_out}). "
            + "This should not be possible! Please, contact a developer!"
        )

        if isinstance(atoms_dict, dict):
            atoms_dict = [atoms_dict]

        atoms = []
        for atom in atoms_dict:
            atoms.append(
                {
                    "element": atom["@name"],
                    "mass": self.atomic_species[atom["@name"]]["mass"],
                    "pseudo_file": self.atomic_species[atom["@name"]]["pseudo_file"],
                    "position_bohr": np.fromstring(atom["#text"], sep=" ", dtype=np.float64),
                    "position_hartree": np.fromstring(atom["#text"], sep=" ", dtype=np.float64),
                    "position_meter": np.fromstring(atom["#text"], sep=" ", dtype=np.float64) * dopyqo.BOHR_TO_METER,
                    "position_angstrom": np.fromstring(atom["#text"], sep=" ", dtype=np.float64) * dopyqo.BOHR_TO_ANGSTROM,
                    "atomic_number": dopyqo.elements_to_atomic_number[atom["@name"]],
                }
            )
        self.atoms = atoms

        # shape (#atoms, 3)
        self.atom_positions_hartree = np.array([atom["position_hartree"] for atom in self.atoms], dtype=np.float64)
        self.atomic_numbers = np.array([atom["atomic_number"] for atom in self.atoms])

        self.atomic_numbers_valence = self.atomic_numbers.copy()
        if pseudopots is not None:
            # Mapping atomic numbers to list of atom positions
            atoms_dict = {num: [] for num in self.atomic_numbers}
            atoms_dict = {key: np.array(val) for key, val in atoms_dict.items()}
            pp_dict = {pp.atomic_number: pp for pp in pseudopots}

            # Checking for duplicate PPs
            atomic_nums_pp_duplicates = [
                atomic_num_pp for atomic_num_pp, count_val in Counter([pp.atomic_number for pp in pseudopots]).items() if count_val > 1
            ]
            atomic_el_pp_duplicates = [dopyqo.elements_to_atomic_number[x] for x in atomic_nums_pp_duplicates]
            assert len(atomic_nums_pp_duplicates) == 0, "More than one pseudopotential given for atoms " + ", ".join(
                str(x) + f" (Z={atomic_nums_pp_duplicates[i]})" for i, x in enumerate(atomic_el_pp_duplicates)
            )

            # Checking for atoms without PP
            atomic_num_wo_pp = [atomic_num for atomic_num in atoms_dict.keys() if atomic_num not in pp_dict.keys()]
            atomic_el_wo_pp = [dopyqo.elements_to_atomic_number[x] for x in atomic_num_wo_pp]
            assert len(atomic_num_wo_pp) == 0, "No pseudopotentials given for atoms " + ", ".join(
                str(x) + f" (Z={atomic_num_wo_pp[i]})" for i, x in enumerate(atomic_el_wo_pp)
            )

            for i, z in enumerate(self.atomic_numbers):
                self.atomic_numbers_valence[i] = pp_dict[z].z_valence

        assert len(self.atom_positions_hartree) == len(self.atomic_numbers), (
            f"Number of atomic positions ({len(self.atom_positions_hartree)}) is not equal "
            + f"to number of atomic numbers ({len(self.atomic_numbers)})! "
            + "Both where read from the xml-file."
        )

        self.atom_positions_hartree_mean = np.mean(self.atom_positions_hartree, axis=0)
        self.atom_center_of_mass = np.sum(self.atomic_numbers[:, None] * self.atom_positions_hartree, axis=0) / np.sum(self.atomic_numbers)

        # Calculate Lattice vectors in bohr/hartree units from reciprocal lattice vectors
        self.cell_volume_reciprocal = self.b1.dot(np.cross(self.b2, self.b3))
        self.a1 = 2 * np.pi / self.cell_volume_reciprocal * np.cross(self.b2, self.b3)
        self.a2 = 2 * np.pi / self.cell_volume_reciprocal * np.cross(self.b3, self.b1)
        self.a3 = 2 * np.pi / self.cell_volume_reciprocal * np.cross(self.b1, self.b2)
        self.cell_volume = np.abs(self.a1.dot(np.cross(self.a2, self.a3)))
        self.cell_volume_reciprocal = np.abs(self.cell_volume_reciprocal)

        # Read reciprocal lattice vectors from xml file
        # are in units of 2*np.pi/alat_out, NOT in 1/bohr = 1/(5.29177210544e-11 meter)
        reciprocal_cell_xml = self.xml_dict["qes:espresso"]["output"]["basis_set"]["reciprocal_lattice"]
        self.b1_xml = np.fromstring(reciprocal_cell_xml["b1"], sep=" ", dtype=np.float64)
        self.b2_xml = np.fromstring(reciprocal_cell_xml["b2"], sep=" ", dtype=np.float64)
        self.b3_xml = np.fromstring(reciprocal_cell_xml["b3"], sep=" ", dtype=np.float64)

        assert np.allclose(
            np.array([self.b1, self.b2, self.b3]), np.array([self.b1_xml, self.b2_xml, self.b3_xml]) * (2 * np.pi) / alat_out
        ), f"Reciprocal lattice vectors in {output_xml} do not match the given reciprocal lattice vectors b1, b2, b3 (possibly from dat/hdf5 file)"

        # Read lattice vectors from xml file
        cell = self.xml_dict["qes:espresso"]["input"]["atomic_structure"]["cell"]
        a1 = np.fromstring(cell["a1"], sep=" ", dtype=np.float64)
        a2 = np.fromstring(cell["a2"], sep=" ", dtype=np.float64)
        a3 = np.fromstring(cell["a3"], sep=" ", dtype=np.float64)

        assert np.allclose(
            np.array([self.a1, self.a2, self.a3]), np.array([a1, a2, a3])
        ), f"Lattice vectors in {output_xml} (\n{np.array([a1, a2, a3])}\n) do not match the given reciprocal lattice vectors b1, b2, b3 (possibly from dat/hdf5 file) and their corresponding lattice vectors (\n{np.array([self.a1, self.a2, self.a3])}\n)"
        self.a = np.array([self.a1, self.a2, self.a3])  # a1 in first row, a2 in second row, a3 in third row.

        ks_energies_entry = self.xml_dict["qes:espresso"]["output"]["band_structure"]["ks_energies"]
        if self.n_kpoints == 1:
            self.ks_energies = np.array([float(x) for x in ks_energies_entry["eigenvalues"]["#text"].split()])
        else:
            self.ks_energies = [np.array([float(x) for x in y["eigenvalues"]["#text"].split()]) for y in ks_energies_entry][self.kpoint_idx]

        # TODO: Define nelec tuple of number of electrons for spin-up and spin-down
        self.nelec = int(float(self.xml_dict["qes:espresso"]["output"]["band_structure"]["nelec"]))

        if self.nelec % 2 == 1:
            assert self.spin == 2, (
                f"The number of electrons ({self.nelec}) is odd but a spin-unpolarized DFT calculation was performed! "
                + "Odd number of electrons are only supported for spin-polarized DFT calculations!"
            )

        if self.n_kpoints == 1:
            self.occupations = np.array([float(x) for x in ks_energies_entry["occupations"]["#text"].split()])
        else:
            self.occupations = [np.array([float(x) for x in y["occupations"]["#text"].split()]) for y in ks_energies_entry][self.kpoint_idx]
        self.occupations_xml = self.occupations.copy()

        # Negative occupations are non-physical and are filtered out
        # but keeping the sum of occupations invariant by adding the sum
        # of negative occupations to the largest occupation
        # This hopefully does not change the physics
        sum_of_neg = np.sum(self.occupations[self.occupations < 0.0])
        self.occupations[self.occupations < 0.0] = 0.0
        self.occupations[np.argmax(self.occupations)] += sum_of_neg

        # Same for occupations greater than 1
        sum_of_gt1 = np.sum(self.occupations[self.occupations > 1.0] - 1.0)
        self.occupations[self.occupations > 1.0] = 1.0
        self.occupations[np.argmin(self.occupations)] += sum_of_gt1

        # Assert the previous
        assert not any(n < 0 for n in self.occupations), f"The occupations cannot be smaller than 0 but are {self.occupations}."
        assert not any(n > 1 for n in self.occupations), f"The occupations cannot be greater than 1 but are {self.occupations}."

        if self.n_kpoints > 1:
            # For k-point calculations the occupations are define wrt to the Fermi energy. This can result in the fact that
            # two times the sum of occupations does not match the number of electrons.
            # For k-point calculations we set to occupations to binary occupations with the correct sum.
            logging.debug("Calculation with more than one k-point. Setting occupations to binary occupations!")
            self.occupations = np.concatenate(
                (
                    np.ones((self.nelec // 2,), dtype=np.int64),
                    np.zeros(
                        (len(self.occupations) - self.nelec // 2,),
                        dtype=np.int64,
                    ),
                ),
                dtype=np.int64,
            )

        assert np.isclose(self.nelec, occ_sum := np.sum(self.occupations) * 2), (
            f"Two times the occupations ({self.occupations}) do not sum up (sum={occ_sum}) "
            + f"to the number of electrons from the xml-file ({self.nelec})!"
        )
        assert np.isclose(round(np.sum(self.occupations)), np.sum(self.occupations)), (
            f"Sum of occupations ({np.sum(self.occupations)}) is not close to an integer, "
            + f"but matches half the number of electrons ({self.nelec/2}). "
            + "This should not be possible! Please, contact a developer!"
        )

        # If occupations are not only 0 and 1 then we do build the occupations_binary vector from scratch
        # Note: Logic not implementable with for-else since:
        #       "If the loop finishes without executing the break, the else clause executes."
        make_binary = False
        for occ in self.occupations:
            if (not np.isclose(occ, 0.0)) and (not np.isclose(occ, 1.0)):
                make_binary = True
                break

        self.occupations_binary = self.occupations.copy().round().astype(np.int64)
        if make_binary:
            self.occupations_binary = np.concatenate(
                (
                    np.ones((round(np.sum(self.occupations)),), dtype=np.int64),
                    np.zeros(
                        (len(self.occupations) - round(np.sum(self.occupations)),),
                        dtype=np.int64,
                    ),
                ),
                dtype=np.int64,
            )

        assert np.isclose(
            np.sum(self.occupations_binary), np.sum(self.occupations)
        ), "Sum of binary occupations is not equal to the sum of occupations. This should not be possible! Please, contact a developer!"

        self.ks_energies_up = self.ks_energies[:nbnd]
        self.occupations_up = self.occupations[:nbnd]
        self.occupations_binary_up = self.occupations_binary[:nbnd]
        if self.spin == 2:
            self.ks_energies_dw = self.ks_energies[nbnd:]
            self.occupations_dw = self.occupations[nbnd:]
            self.occupations_binary_dw = self.occupations_binary[nbnd:]
        else:
            self.ks_energies_dw = self.ks_energies_up
            self.occupations_dw = self.occupations_up
            self.occupations_binary_dw = self.occupations_binary_up

    @classmethod
    def from_file(
        cls,
        file: str,
        output_xml: str,
        pseudopots: list[dopyqo.Pseudopot] = None,
        kpoint_idx: int | None = None,
    ):
        """Create Wfc object from a .dat or .hdf5 Quantum ESPRESSO output file.
        Calls Wfc.from_dat_file or Wfc.from_hdf5_file depending on the provided file.

        Args:
            file (str): Path to the .dat or .hdf5 file.
            output_xml (str): Path to the XML-file Quantum ESPRESSO outputs.
            pseudopots (list[Pseudopot], optional): List of dopyqo.Pseudopot objects, one object per atom type. Defaults to None.
            kpoint_idx (int | None, optional): k-point index starting from 0. Is set to 0, if None and Quantum ESPRESSO calculation
                                               only involved one k-point. Has to match ik+1, if Quantum ESPRESSO calculation
                                               involved multiple k-points. Defaults to None.

        Raises:
            NotImplementedError: If the provided file is not a .dat and not .hdf5 file.

        Returns:
            Wfc: Wfc object
        """
        if file.endswith("dat"):
            return Wfc.from_dat_file(file, output_xml, pseudopots, kpoint_idx)
        elif file.endswith("hdf5"):
            return Wfc.from_hdf5_file(file, output_xml, pseudopots, kpoint_idx)
        else:
            raise NotImplementedError(f"File extension {file.split('.')[-1]} not supported!")

    @classmethod
    def from_hdf5_file(
        cls,
        hdf5_file: str,
        output_xml: str,
        pseudopots: list[dopyqo.Pseudopot] = None,
        kpoint_idx: int | None = None,
    ):
        """Create Wfc object from a .hdf5 Quantum ESPRESSO output file.

        Args:
            hdf5_file (str): Path to the .hdf5 file.
            output_xml (str): Path to the XML-file Quantum ESPRESSO outputs.
            pseudopots (list[Pseudopot], optional): List of dopyqo.Pseudopot objects, one object per atom type. Defaults to None.
            kpoint_idx (int | None, optional): k-point index starting from 0. Is set to 0, if None and Quantum ESPRESSO calculation
                                               only involved one k-point. Has to match ik+1, if Quantum ESPRESSO calculation
                                               involved multiple k-points. Defaults to None.

        Returns:
            Wfc: Wfc object
        """
        f = h5py.File(hdf5_file, "r")  # Works like a python dictionary
        # HDF5 files contain datasets which have a shape and a dtype attribute.
        # HDF5 files and all containing datasets also contain attributes which
        # can be obtained with f.attrs.keys().

        # The coefficients alternate between the real and imaginary part
        evc_real_imag = np.array(f["evc"])
        mill = np.array(f["MillerIndices"])
        b1 = np.array(f["MillerIndices"].attrs["bg1"])
        b2 = np.array(f["MillerIndices"].attrs["bg2"])
        b3 = np.array(f["MillerIndices"].attrs["bg3"])
        gamma_only = "TRUE" in str(f.attrs["gamma_only"])
        igwx = f.attrs["igwx"]
        ik = f.attrs["ik"]
        ispin = f.attrs["ispin"]
        nbnd = f.attrs["nbnd"]
        ngw = f.attrs["ngw"]
        npol = f.attrs["npol"]
        scale_factor = f.attrs["scale_factor"]
        xk = f.attrs["xk"]
        f.close()

        evc = np.zeros(
            shape=(evc_real_imag.shape[0], evc_real_imag.shape[1] // 2),
            dtype=np.complex128,
        )
        evc = evc_real_imag[:, 0::2] + 1j * evc_real_imag[:, 1::2]

        return cls(
            ik=ik,
            xk=xk,
            ispin=ispin,
            gamma_only=gamma_only,
            scalef=scale_factor,
            ngw=ngw,
            igwx=igwx,
            npol=npol,
            nbnd=nbnd,
            b1=b1,
            b2=b2,
            b3=b3,
            mill=mill,
            evc=evc,
            output_xml=output_xml,
            pseudopots=pseudopots,
            kpoint_idx=kpoint_idx,
        )

    @classmethod
    def from_dat_file(
        cls,
        dat_file: str,
        output_xml: str,
        pseudopots: list[dopyqo.Pseudopot] = None,
        kpoint_idx: int | None = None,
    ):
        """Create Wfc object from a .dat Quantum ESPRESSO output file.

        Args:
            dat_file (str): Path to the .dat file.
            output_xml (str): Path to the XML-file Quantum ESPRESSO outputs.
            pseudopots (list[Pseudopot], optional): List of dopyqo.Pseudopot objects, one object per atom type. Defaults to None.
            kpoint_idx (int | None, optional): k-point index starting from 0. Is set to 0, if None and Quantum ESPRESSO calculation
                                               only involved one k-point. Has to match ik+1, if Quantum ESPRESSO calculation
                                               involved multiple k-points. Defaults to None.

        Returns:
            Wfc: Wfc object
        """
        # Taken from https://mattermodeling.stackexchange.com/questions/9149/how-to-read-qes-wfc-dat-files-with-python
        # INTEGER :: ik
        # !! k-point index (1 to number of k-points)
        # REAL(8) :: xk(3)
        # !! k-point coordinates
        # INTEGER :: ispin
        # !! spin index for LSDA case: ispin=1 for spin-up, ispin=2 for spin-down
        # !! for unpolarized or non-colinear cases, ispin=1 always
        # LOGICAL :: gamma_only
        # !! if .true. write or read only half of the plane waves
        # REAL(8) :: scalef
        # !! scale factor applied to wavefunctions
        # INTEGER :: ngw
        # !! number of plane waves (PW)
        # INTEGER :: igwx
        # !! max number of PW (may be larger than ngw, not sure why)
        # INTEGER :: npol
        # !! number of spin states for PWs: 2 for non-colinear case, 1 otherwise
        # INTEGER :: nbnd
        # !! number of wavefunctions
        # REAL(8) :: b1(3), b2(3), b3(3)
        # !! primitive reciprocal lattice vectors
        # INTEGER :: mill(3,igwx)
        # !! miller indices: h=mill(1,i), k=mill(2,i), l=mill(3,i)
        # !! the i-th PW has wave vector (k+G)(:)=xk(:)+h*b1(:)+k*b2(:)+ l*b3(:)
        # COMPLEX(8) :: evc(npol*igwx,nbnd)
        # !! wave functions in the PW basis set
        # !! The first index runs on PW components,
        # !! the second index runs on band states.
        # !! For non-colinear case, each PW has a spin component
        # !! first  igwx components have PW with   up spin,
        # !! second igwx components have PW with down spin
        with open(dat_file, "rb") as f:
            # Moves the cursor 4 bytes to the right
            f.seek(4)

            ik = np.fromfile(f, dtype="int32", count=1)[0]
            xk = np.fromfile(f, dtype="float64", count=3)
            ispin = np.fromfile(f, dtype="int32", count=1)[0]
            gamma_only = bool(np.fromfile(f, dtype="int32", count=1)[0])
            scalef = np.fromfile(f, dtype="float64", count=1)[0]

            # Move the cursor 8 byte to the right
            f.seek(8, 1)

            ngw = np.fromfile(f, dtype="int32", count=1)[0]
            igwx = np.fromfile(f, dtype="int32", count=1)[0]
            npol = np.fromfile(f, dtype="int32", count=1)[0]
            nbnd = np.fromfile(f, dtype="int32", count=1)[0]

            # Move the cursor 8 byte to the right
            f.seek(8, 1)

            b1 = np.fromfile(f, dtype="float64", count=3)
            b2 = np.fromfile(f, dtype="float64", count=3)
            b3 = np.fromfile(f, dtype="float64", count=3)

            f.seek(8, 1)

            mill = np.fromfile(f, dtype="int32", count=3 * igwx)
            mill = mill.reshape((igwx, 3))

            evc = np.zeros((nbnd, npol * igwx), dtype="complex128")

            f.seek(8, 1)
            for i in range(nbnd):
                evc[i, :] = np.fromfile(f, dtype="complex128", count=npol * igwx)
                f.seek(8, 1)

        return cls(
            ik=ik,
            xk=xk,
            ispin=ispin,
            gamma_only=gamma_only,
            scalef=scalef,
            ngw=ngw,
            igwx=igwx,
            npol=npol,
            nbnd=nbnd,
            b1=b1,
            b2=b2,
            b3=b3,
            mill=mill,
            evc=evc,
            output_xml=output_xml,
            pseudopots=pseudopots,
            kpoint_idx=kpoint_idx,
        )

    @classmethod
    def from_file_all_kpoints(
        cls,
        files: list[str],
        output_xml: str,
        pseudopots: list[dopyqo.Pseudopot] = None,
        verbose: bool = False,
    ):
        """Create Wfc object from a list of .dat or .hdf5 Quantum ESPRESSO output files and return a list of Wfc objects.
        Calls Wfc.from_file for each file in the list of files.

        Args:
            files (list[str]): List of .dat or .hdf5 files
            output_xml (str): Path to the XML-file Quantum ESPRESSO outputs.
            pseudopots (list[dopyqo.Pseudopot], optional): List of dopyqo.Pseudopot objects, one object per atom type. Defaults to None.

        Returns:
            list[Wfc]: List of Wfc objects
        """
        assert isinstance(files, list), f"Expected list of files but got {files.__class__.__name__}!"

        with open(output_xml, encoding="utf-8") as file:
            xml_dict = xmltodict.parse(file.read())
        # Read number of k-points
        n_kpoints = int(xml_dict["qes:espresso"]["output"]["band_structure"]["nks"])

        # Create Wfc object for every k-point
        wfc_obj_list = []
        for i in range(n_kpoints):
            if verbose:
                print(f"Loading k-point {i+1}/{n_kpoints}...", end="\r", flush=True)
            wfc_obj_list.append(
                Wfc.from_file(
                    files[i],
                    output_xml,
                    pseudopots,
                    kpoint_idx=i,
                )
            )
        return wfc_obj_list

    def get_overlaps(self, orbitals=None):
        if orbitals is None:
            overlaps = np.einsum("ij, kj -> ik", self.evc.conj(), self.evc)
        else:
            overlaps = np.einsum("ij, kj -> ik", self.evc[orbitals].conj(), self.evc[orbitals])

        return overlaps

    def check_norm(self):
        assert np.allclose(self.get_overlaps(), np.identity(self.evc.shape[0])), "Overlap matrix is not diagonal (Wavefunctions are not orthonormal)!"

        logging.info("Overlap matrix is diagonal (Wavefunctions are orthonormal)!")

    def get_orbitals_by_index(self, indices: list | np.ndarray | None = None, binary_occupations=True):
        if indices is None:
            indices = list(range(len(self.occupations)))

        indices = indices.copy()

        if binary_occupations:
            occupations = self.occupations_binary[indices]
        else:
            occupations = self.occupations[indices]

        c_ip_orbitals = self.evc[indices]

        if len(indices) == 1:
            indices.append(indices[0])

        warn_msg = "The selected orbitals for the active space are all fully occupied or unoccupied!"
        occ_str = " ".join([(f"|{val}" if i == indices[0] else f"{val}|" if i == indices[1] else f"{val}") for i, val in enumerate(occupations)])
        warn_msg += f" You selected the following active space: {occ_str}"
        if np.any(occupations == 2):
            if not np.any(occupations == 1):
                warn(warn_msg)
        elif np.all(occupations == 0):
            warn(warn_msg)
        else:
            if not np.any(occupations == 0):
                warn(warn_msg)

        return occupations, c_ip_orbitals

    # Copied and adjusted from the pennylane function (e.g. pennylane version 0.40.0)
    # https://docs.pennylane.ai/en/stable/code/api/pennylane.qchem.active_space.html
    def active_space(self, active_electrons: int, active_orbitals: int) -> tuple[list[int], list[int]]:
        if self.spin == 1:
            mult = 1
        elif self.spin == 2:
            mult = np.abs(np.sum(self.occupations_up) - np.sum(self.occupations_dw))
            assert np.isclose(round(mult), mult), f"Calculated spin multiplicity ({mult}) is not close to an integer!"
        else:
            raise ValueError(f"Spin is neither 1 nor 2, but {self.spin}")
        if active_electrons is None:
            ncore_orbs = 0
            core = []
        else:
            if active_electrons <= 0:
                raise ValueError(f"The number of active electrons ({active_electrons}) " f"has to be greater than 0.")

            if active_electrons > self.nelec:
                raise ValueError(
                    f"The number of active electrons ({active_electrons}) "
                    f"can not be greater than the total "
                    f"number of electrons ({self.nelec})."
                )

            if active_electrons < mult - 1:
                raise ValueError(
                    f"For a reference state with multiplicity {mult}, "
                    f"the number of active electrons ({active_electrons}) should be "
                    f"greater than or equal to {mult - 1}."
                )

            if mult % 2 == 1:
                if active_electrons % 2 != 0:
                    raise ValueError(
                        f"For a reference state with multiplicity {mult}, " f"the number of active electrons ({active_electrons}) should be even."
                    )
            else:
                if active_electrons % 2 != 1:
                    raise ValueError(
                        f"For a reference state with multiplicity {mult}, " f"the number of active electrons ({active_electrons}) should be odd."
                    )

            ncore_orbs = (self.nelec - active_electrons) // 2
            core = list(range(ncore_orbs))

        if active_orbitals is None:
            active = list(range(ncore_orbs, self.nbnd))
        else:
            if active_orbitals <= 0:
                raise ValueError(f"The number of active orbitals ({active_orbitals}) " f"has to be greater than 0.")

            if ncore_orbs + active_orbitals > self.nbnd:
                raise ValueError(
                    f"The number of core ({ncore_orbs}) + active orbitals ({active_orbitals}) cannot "
                    f"be greater than the total number of orbitals ({self.nbnd})"
                )

            homo = (self.nelec + mult - 1) / 2
            if ncore_orbs + active_orbitals <= homo:
                raise ValueError(f"For n_active_orbitals={active_orbitals}, there are no virtual orbitals " f"in the active space.")

            active = list(range(ncore_orbs, ncore_orbs + active_orbitals))

        return core, active

    def set_atom_positions(self, atom_positions: np.ndarray, unit: dopyqo.Unit):
        """Change the atom positions

        Args:
            atom_positions (np.ndarray): New atom positions. Shape (# of atoms, 3).
            unit (dopyqo.Unit): Unit in which the new atom positions are given

        Raises:
            ValueError: If the unit is not supported.
        """
        assert len(atom_positions) == len(self.atomic_numbers), (
            f"Number of new atomic positions ({len(atom_positions)}) is not equal " + f"to number of atomic numbers ({len(self.atomic_numbers)})!"
        )
        assert atom_positions.shape[1] == 3, (
            "New atomic positions should define x,y, and z coordinates " + f"but they have {atom_positions.shape[1]} coordinates!"
        )

        atom_postions_hartree = None
        allowed_units_str = "\n\t".join(dopyqo.ALLOWED_UNITS)
        match unit:
            case dopyqo.Unit.HARTREE:
                atom_postions_hartree = atom_positions
            case dopyqo.Unit.ANGSTROM:
                atom_postions_hartree = atom_positions / dopyqo.BOHR_TO_ANGSTROM
            case dopyqo.Unit.METER:
                atom_postions_hartree = atom_positions / dopyqo.BOHR_TO_METER
            case dopyqo.Unit.ALAT:
                atom_postions_hartree = atom_positions * self.alat
            case dopyqo.Unit.CRYSTAL:
                atom_postions_hartree = np.array([np.sum(pos * self.a, axis=0) for pos in atom_positions])
            case _:
                raise ValueError(f"Unit {unit} not supported for setting atom positions! Supported units are:\n\t{allowed_units_str}\n")

        atoms = []
        for i, atom in enumerate(self.atoms):
            atom["position_hartree"] = atom_postions_hartree[i]
            atom["position_bohr"] = atom_postions_hartree[i]
            atom["position_meter"] = atom_postions_hartree[i] * dopyqo.BOHR_TO_METER
            atom["position_angstrom"] = atom_postions_hartree[i] * dopyqo.BOHR_TO_ANGSTROM
            atoms.append(atom)
        self.atoms = atoms
        # shape (#atoms, 3)
        self.atom_positions_hartree = np.array([atom["position_hartree"] for atom in self.atoms], dtype=np.float64)

        self.atom_positions_hartree_mean = np.mean(self.atom_positions_hartree, axis=0)
        self.atom_center_of_mass = np.sum(self.atomic_numbers[:, None] * self.atom_positions_hartree, axis=0) / np.sum(self.atomic_numbers)

    def set_lattice_vectors(self, lattice_vectors: np.ndarray, unit: dopyqo.Unit):
        """Change the lattice vectors

        Args:
            lattice_vectors (np.ndarray): New lattice vectors. Shape (3, 3). Each row is one lattice vector.
            unit (dopyqo.Unit):  Unit in which the new atom positions are given

        Raises:
            ValueError:  If the unit is not supported.
        """
        if lattice_vectors.shape != (3, 3):
            raise ValueError(f"Lattice vectors are invalid. Expected a list of array of shape (3, 3) but got {lattice_vectors.shape}.")

        lattice_vectors_hartree = None
        allowed_units_lattice = dopyqo.ALLOWED_UNITS_LATTICE
        allowed_units_lattice_str = "\n\t".join(allowed_units_lattice)
        match unit:
            case dopyqo.Unit.HARTREE:
                lattice_vectors_hartree = lattice_vectors
            case dopyqo.Unit.ANGSTROM:
                lattice_vectors_hartree = lattice_vectors / dopyqo.BOHR_TO_ANGSTROM
            case dopyqo.Unit.METER:
                lattice_vectors_hartree = lattice_vectors / dopyqo.BOHR_TO_METER
            case _:
                raise ValueError(f"Unit {unit} not supported for setting lattice vectors! Supported units are:\n\t{allowed_units_lattice_str}\n")

        self.a1 = lattice_vectors_hartree[0]
        self.a2 = lattice_vectors_hartree[1]
        self.a3 = lattice_vectors_hartree[2]
        self.cell_volume = np.abs(self.a1.dot(np.cross(self.a2, self.a3)))

        self.a = np.array([self.a1, self.a2, self.a3])
        self.b = np.array([self.b1, self.b2, self.b3])

        self.b1 = 2 * np.pi / self.cell_volume * np.cross(self.a2, self.a3)
        self.b2 = 2 * np.pi / self.cell_volume * np.cross(self.a3, self.a1)
        self.b3 = 2 * np.pi / self.cell_volume * np.cross(self.a1, self.a2)
        self.cell_volume_reciprocal = np.abs(self.b1.dot(np.cross(self.b2, self.b3)))

        # See https://www.quantum-espresso.org/Doc/INPUT_PW.html CARD: CELL_PARAMETERS description
        self.alat = np.sqrt(self.a1 @ self.a1)

    def wigner_seitz_cell(self) -> list[list[np.ndarray]]:
        """Calculate all facets of the Wigner-Seitz cell. See dopyqo.wfc.plot_wigner_seitz_cell to plot the return values of this function.

        Returns:
            list[list[np.ndarray]]: List of facets
        """
        facets = wigner_seitz_cell(self.a1, self.a2, self.a3)
        return facets

    def wigner_seitz_in_radius(self) -> float:
        """Calculate the inner radius of the Wigner-Seitz cell.

        Returns:
            float: Inner radius of the Wigner-Seitz cell
        """
        facets = wigner_seitz_cell(self.a1, self.a2, self.a3)
        r_in = wigner_seitz_in_radius(facets)
        return r_in

    def to_qe_input(self, filename: str, prefix: str, outdir: str = "./", pseudo_dir: str | None = None):
        """
        Generates a Quantum ESPRESSO (QE) input file for a self-consistent field (SCF) calculation.
        Author: Erik Hansen

        Args:
            filename (str): Full path and filename of the input file to create
            prefix (str): Prefix for the QE calculation
            outdir (str): Output directory specified in the input file ('outdir')
            pseudo_dir (str | None, optional): Directory with the pseudopotential files specified in the input file ('pseudo_dir').
                                               Not written to input file if None. Defaults to None.
        """
        skip_occ = False
        if self.occupations_scf == "from_input":
            print(
                f"{ORANGE}Warning: 'occupations' in original QE input file was set to 'from_input'. "
                + f"This is currently not supported. 'occupations' will be skipped in the generated QE input file.{RESET_COLOR}"
            )
            skip_occ = True

        folder_path = os.path.join(*os.path.split(filename)[:-1])
        if len(folder_path) > 0:
            os.makedirs(folder_path, exist_ok=True)
        # Write the new .in file
        with open(filename, "w", encoding="utf-8") as f:
            f.write("&CONTROL\n")
            f.write("  calculation = 'scf',\n")
            f.write("  forc_conv_thr = 0.001,\n")  # fi problems, maby extract from xml file and use this one
            f.write(f"  prefix = '{prefix}',\n")
            f.write("  verbosity = 'high',\n")
            f.write(f"  outdir = '{outdir}',\n")
            if pseudo_dir is not None:
                f.write(f"  pseudo_dir = '{pseudo_dir}',\n")
            f.write("/\n\n")
            f.write("&SYSTEM\n")
            f.write(f"  ecutwfc = {self.ecutwfc*2},\n")  # Hartree to Ry
            f.write(f"  ecutrho = {self.ecutrho*2},\n")  # Hartree to Ry
            f.write(f"  ibrav = 0,\n")
            f.write(f"  nat = {len(self.atom_positions_hartree)},\n")
            f.write(f"  ntyp = {len(self.atomic_species)},\n")  # Extract number of species names
            f.write(f"  nbnd = {self.nbnd},\n")  # TODO: a formula to calculate it might be needed
            if not skip_occ:
                f.write(f"  occupations = '{self.occupations_scf}',\n")
                if self.occupations_scf == "smearing":
                    f.write(f"  degauss = {self.degauss * 2.0},\n")  # Hartree to Ry
                    f.write(f"  smearing = '{self.smearing}',\n")
            f.write("/\n\n")
            f.write("&ELECTRONS\n")
            f.write(f"  conv_thr = {self.conv_thr * 2.0},\n")  # Hartree to Ry
            f.write(f"  electron_maxstep = {self.electron_maxstep},\n")
            f.write(f"  mixing_beta = {self.mixing_beta},\n")
            f.write("/\n\n")
            f.write("&IONS\n")
            f.write("/\n\n")
            f.write("&CELL\n")
            f.write("/\n\n")
            f.write("CELL_PARAMETERS bohr\n")
            for vector in self.a:
                f.write(f"  {vector[0]} {vector[1]} {vector[2]}\n")
            f.write("\n")
            f.write("ATOMIC_SPECIES\n")
            for name, vals in self.atomic_species.items():
                mass = vals["mass"]
                pseudo_file = vals["pseudo_file"]
                f.write(f"  {name} {mass} {pseudo_file}\n")
            f.write("\n")
            f.write("ATOMIC_POSITIONS bohr\n")
            for atom in self.atoms:
                name = atom["element"]
                pos = atom["position_hartree"]
                f.write(f"  {name} {pos[0]} {pos[1]} {pos[2]}\n")
            f.write("\n")
            f.write("K_POINTS gamma\n")
        logging.info("QE SCF input file written to: %s", filename)

    def plot_real_space(
        self,
        isosurfaces: int | list[float],
        band_idc: list | np.ndarray | None = None,
        plot_abs: bool = True,
        plot_lattice_vectors: bool = True,
        extend_data: bool = False,
        plotter: pv.Plotter = None,
        **kwargs,
    ) -> pv.Plotter:
        """Plot the real-space representation of the Kohn-Sham orbitals

        Args:
            isosurfaces (int | list[float]): Number of isosurfaces, if int. Values of isosurfaces if list of floats
            band_idc (list | np.ndarray | None, optional): Indices of Kohn-Sham orbitals to plot. Plots all Kohn-Sham orbitals if None. Defaults to None.
            plot_abs (bool, optional): If the absolute values of the Kohn-Sham orbitals are plotted. Defaults to True.
            plot_lattice_vectors (bool, optional): If the lattice vectors are plotted as arrows. Defaults to True.
            extend_data (bool, optional): If the orbitals are also plotted for all neighbouring cells besides the computational cell. Defaults to False.
            plotter (pv.Plotter, optional): PyVista plotter object used for plotting. A new one is created if None. Defaults to None.
            kwargs: Arguments passed to dopyqo.plotting.plot_3d_data

        Returns:
            pv.Plotter: PyVista plotter object
        """
        if band_idc is None:
            band_idc = list(range(self.nbnd))

        # Real space wavecfunctions
        max_min = self.fft_grid
        mill = self.mill
        tot_elec = self.nelec
        occupations_active, c_ip_active = self.get_orbitals_by_index(band_idc, binary_occupations=True)
        c_ip = c_ip_active

        return self.plot_real_space_from_c_ip(
            c_ip=c_ip,
            plot_abs=plot_abs,
            plot_lattice_vectors=plot_lattice_vectors,
            extend_data=extend_data,
            isosurfaces=isosurfaces,
            plotter=plotter,
            **kwargs,
        )

    def plot_real_space_from_c_ip(
        self,
        c_ip: np.ndarray,
        isosurfaces: int | list[float],
        plot_abs: bool = True,
        plot_lattice_vectors: bool = True,
        extend_data: bool = False,
        extend_atoms: bool = False,
        plotter: pv.Plotter | None = None,
        **kwargs,
    ) -> pv.Plotter:
        """Plot the real-space representation of the Kohn-Sham orbitals, given their coefficients in the plane wave basis

        Args:
            c_ip (np.ndarray): Numpy array of plane wave coefficients of shape (nbnd, npw), where nbnd is the number of bands and npw is the number of plane waves
            isosurfaces (int | list[float]): Number of isosurfaces, if int. Values of isosurfaces if list of floats
            plot_abs (bool, optional): If the absolute values of the Kohn-Sham orbitals are plotted. Defaults to True.
            plot_lattice_vectors (bool, optional): If the lattice vectors are plotted as arrows. Defaults to True.
            extend_data (bool, optional): If the orbitals are also plotted for all neighbouring cells besides the computational cell. Defaults to False.
            extend_atoms (bool, optional): If the atoms are also plotted for all neighbouring cells besides the computational cell. Defaults to False.
            plotter (pv.Plotter | None, optional): PyVista plotter object used for plotting. A new one is created if None. Defaults to None.
            kwargs: Arguments passed to dopyqo.plotting.plot_3d_data

        Returns:
            pv.Plotter: PyVista plotter object
        """
        max_min = self.fft_grid
        mill = self.mill
        nbands = c_ip.shape[0]
        c_ip_array = np.zeros(
            (
                nbands,
                *max_min,
            ),
            dtype=c_ip.dtype,
        )
        #
        # Set \psi_i(p) on given grid points
        assert mill.shape[0] == c_ip.shape[1]
        for idx, mill_idx in enumerate(mill):
            x, y, z = mill_idx
            i, j, k = (
                x + max_min[0] // 2,
                y + max_min[1] // 2,
                z + max_min[2] // 2,
            )
            c_ip_array[:, i, j, k] = c_ip[:, idx]
        #
        psi_r = []
        for i in range(nbands):
            c_ip_shifted = np.fft.ifftshift(c_ip_array[i])
            psi_r_i = np.fft.ifftn(c_ip_shifted) * np.sqrt(np.prod(max_min))
            # psi_r_i = np.fft.fftshift(psi_r_i)
            psi_r.append(psi_r_i)
        psi_r = np.array(psi_r)
        if plot_abs:
            psi_r = np.abs(psi_r) ** 2

        nx, ny, nz = max_min
        origin = np.zeros(3)
        # Looked good for diamond but yielded wrong plot for H2
        coordinates = (
            np.arange(nx)[:, None] * self.a[0] / nx
            + np.arange(ny)[:, None, None] * self.a[1] / ny
            + np.arange(nz)[:, None, None, None] * self.a[2] / nz
            + origin
        )  # .reshape((-1, 3))
        # Following two are equivalent and yield correct plot for H2 compared to the XSF-files
        coordinates = (
            np.arange(nx)[:, None, None, None] * self.a[0] / nx
            + np.arange(ny)[:, None, None] * self.a[1] / ny
            + np.arange(nz)[:, None] * self.a[2] / nz
            + origin
        )  # .reshape((-1, 3))
        #
        # coordinates = (
        #     np.arange(nz)[:, None] * self.a[2] / nz
        #     + np.arange(ny)[:, None, None] * self.a[1] / ny
        #     + np.arange(nx)[:, None, None, None] * self.a[0] / nx
        #     + origin
        # )  # .reshape((-1, 3))

        return self.plot_real_space_from_3d_data(
            coordinates=coordinates,
            psi_r=psi_r,
            plot_lattice_vectors=plot_lattice_vectors,
            extend_data=extend_data,
            extend_atoms=extend_atoms,
            isosurfaces=isosurfaces,
            plotter=plotter,
            **kwargs,
        )

    def plot_real_space_from_3d_data(
        self,
        coordinates: np.ndarray,
        isosurfaces: int | list[float],
        psi_r: list[np.ndarray],
        plot_lattice_vectors: bool = True,
        extend_data: bool = False,
        extend_atoms: bool = False,
        plotter: pv.Plotter | None = None,
        **kwargs,
    ) -> pv.Plotter:
        """_summary_

        Args:
            coordinates (np.ndarray): Numpy array of coordinates of shape (nx,ny,nz,3) or of shape (nx*ny*nz,3)
            isosurfaces (int | list[float]): Number of isosurfaces, if int. Values of isosurfaces if list of floats
            psi_r (list[np.ndarray]): List of numpy arrays. Each numpy array represents scalar field data of shape (nx,ny,nz) or of shape (nx*ny*nz)
            plot_lattice_vectors (bool, optional): If the lattice vectors are plotted as arrows. Defaults to True.
            extend_data (bool, optional): If the orbitals are also plotted for all neighbouring cells besides the computational cell. Defaults to False.
            extend_atoms (bool, optional): If the atoms are also plotted for all neighbouring cells besides the computational cell. Defaults to False.
            plotter (pv.Plotter | None, optional): _description_. Defaults to None.
            kwargs: Arguments passed to dopyqo.plotting.plot_3d_data

        Returns:
            pv.Plotter: PyVista plotter object
        """
        max_min = psi_r[0].shape

        origin = np.zeros(3)

        ################################
        atom_positions_hartree = self.atom_positions_hartree.copy()
        # # atom positions in crystal coordinates
        # atom_positions_crystal = np.array([np.linalg.solve(self.a.T, x) for x in atom_positions_hartree])
        # # atom positions in datagrid coordinates
        # atom_positions_crystal = np.array([x * (max_min) for x in atom_positions_crystal])

        # Move atom positions into home cell spanned by lattice vectors starting from origin
        atom_positions_hartree_in_home_cell = []
        for pos_tmp in atom_positions_hartree:
            pos_new_tmp = pos_tmp.copy()
            inside, coeffs = is_point_inside_parallelepiped(pos_tmp - origin, self.a[0], self.a[1], self.a[2])
            if not inside:
                move_coeffs = -np.sign(coeffs) * np.ceil(np.abs(np.minimum(coeffs, 0)))
                pos_new_tmp = pos_tmp - origin + self.a.T @ move_coeffs

                inside_new, _ = is_point_inside_parallelepiped(pos_new_tmp, self.a[0], self.a[1], self.a[2])
                if not inside_new:
                    print(
                        f"{ORANGE}Plot warning: Could not move atom at position {pos_tmp} into home cell. This should never happen. Please, contact a developer!"
                        + f"\n\tMore information: {inside=} | {coeffs=} | {move_coeffs=} | {pos_new_tmp=} | {self.a=}{RESET_COLOR}"
                    )
                    pos_new_tmp = pos_tmp
            atom_positions_hartree_in_home_cell.append(pos_new_tmp)
        atom_positions_hartree = np.array(atom_positions_hartree_in_home_cell)

        atom_labels = [f"{self.atoms[idx_tmp]['element']} [{idx_tmp}]" for idx_tmp, _ in enumerate(atom_positions_hartree)]

        if extend_atoms:
            n_max_x = 2
            n_max_y = 2
            n_max_z = 2
            ranges = itertools.product(
                range(-1, n_max_x + 1),
                range(-1, n_max_y + 1),
                range(-1, n_max_z + 1),
            )
        else:
            n_max_x = 1
            n_max_y = 1
            n_max_z = 1
            ranges = itertools.product(
                range(-n_max_x - 0, n_max_x + 1),
                range(-n_max_y - 0, n_max_y + 1),
                range(-n_max_z - 0, n_max_z + 1),
            )

        t_vecs_unordered = [[np.dot(self.a.T, np.array([nx, ny, nz])), [nx, ny, nz]] for nx, ny, nz in ranges]
        t_vecs = sorted(t_vecs_unordered, key=lambda x: np.linalg.norm(x[0], ord=2))
        sphere_centers_hartree = []
        sphere_labels = []
        for t_vec, (nx, ny, nz) in t_vecs:
            for idx_tmp, pos in enumerate(atom_positions_hartree):
                sphere_centers_hartree.append(pos + t_vec)
                sphere_labels.append(f"{self.atoms[idx_tmp]['element']} [{idx_tmp}]")
        sphere_centers_hartree_filtered = []
        sphere_labels_filtered = []
        for x, label in zip(sphere_centers_hartree, sphere_labels):
            inside, _ = is_point_inside_parallelepiped(x - origin, self.a[0] * n_max_x, self.a[1] * n_max_y, self.a[2] * n_max_z, tol=1e-2)
            if inside:
                sphere_centers_hartree_filtered.append(x)
                sphere_labels_filtered.append(label)

        sphere_centers_hartree = sphere_centers_hartree_filtered
        sphere_labels = sphere_labels_filtered

        sphere_centers_hartree = np.array(sphere_centers_hartree)

        lattice_vectors_meshes = [pv.Arrow(direction=v, shaft_radius=2e-2, tip_radius=5e-2, scale="auto") for v in self.a]
        meshes_tmp = []
        if plot_lattice_vectors:
            meshes_tmp.extend(lattice_vectors_meshes)

        for data in psi_r:
            if not np.allclose(data.imag, 0.0):
                print(f"{ORANGE}Plot warning: Data is complex. Only real part is plotted, imaginary part is discarded!{RESET_COLOR}")
            data = data.real
            if "meshes" in kwargs:
                kwargs["meshes"].extend(meshes_tmp)
            else:
                kwargs["meshes"] = meshes_tmp
            plotter = plotting.plot_3d_data(
                coordinates=coordinates,
                data=data,
                grid_vectors=self.a.copy(),
                plotter=plotter,
                atom_positions=sphere_centers_hartree,
                atom_labels=sphere_labels,
                isosurfaces=isosurfaces,
                extend_data=extend_data,
                **kwargs,
            )
        return plotter


def runQE(
    input_file: str,
    num_cpus: int = 1,
    nk: int = 1,
    nb: int = 1,
    nt: int = 1,
):
    """
    Executes a Quantum Espresso (QE) simulation by generating the necessary input files,
    running the QE executable, and managing the working directory.
    Author: Erik Hansen

    See https://www.quantum-espresso.org/Doc/user_guide/node20.html
    for arguments nk, nb, nt.

    Side Effects:
        - Creates a new folder for the QE run and generates input files.
        - Executes the Quantum Espresso `pw.x` command using `mpirun`.
        - Changes the working directory to the QE run folder and back to the original path.
        - Writes output to a file named `<prefix_of_QE_current>_scf.out`.
    Notes:
        - Ensure that Quantum Espresso is installed and accessible via the `pw.x` command.
        - The `mpirun` command assumes 10 cores are available for parallel execution.

    Args:
        input_file (str): Full path and filename of the input file.
        num_cpus (int): Number of MPI processes
        nk (int): number of processes for k-point parallelization
        nb (int): number of processes for KS-band parallelization
        nt (int): number of processes for plane wave parallelization
    """
    # current_path = os.path.dirname(os.path.abspath(__file__))
    current_path = os.getcwd()
    logging.info("Changing directory...")

    folder_path = os.path.join(*os.path.split(input_file)[:-1])
    if len(folder_path) > 0:
        os.chdir(folder_path)
    input_file = os.path.split(input_file)[-1]
    if input_file.endswith(".in"):
        output_file = f"{input_file[:-3]}.out"
    else:
        output_file = f"{input_file}.out"
    print("Running Quantum Espresso...")
    command = f"mpirun -n {num_cpus} --bind-to core pw.x -nk {nk} -nb {nb} -nt {nt} -in {input_file} > {output_file}"
    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    output_out, _output_err = process.communicate()
    if output_out is not None and len(output_out) > 0:
        print(f"{RED}Quantum ESPRESSO error output:{SOFT_MAGENTA}\n{textwrap.indent(output_out, '    ')}{RESET_COLOR}")
    process.wait()  # Wait for the process to complete
    # Check for crash
    if "CRASH" in os.listdir():
        # Print also in cases where CRASH file cannot be read
        print(
            f"{RED}Quantum ESPRESSO error: QE calculation crashed.{RESET_COLOR}",
            file=sys.stderr,
        )
        with open("CRASH", "r", encoding="UTF-8") as file:
            contents = file.read()
        print(
            f"{RED}Content of the 'CRASH' file in\n{os.getcwd()}:\n{ORANGE}{contents}{RESET_COLOR}",
            file=sys.stderr,
        )
        os.chdir(current_path)
        sys.exit(1)
    # Exit the run folder
    os.chdir(current_path)
    print(f"{GREEN}Done!{RESET_COLOR}")


# Adjusted from pymatgen.core.lattice.get_wigner_seitz_cell
# https://github.com/materialsproject/pymatgen/blob/v2025.5.28/src/pymatgen/core/lattice.py#L1304-L1326
def wigner_seitz_cell(a1: np.ndarray, a2: np.ndarray, a3: np.ndarray) -> list[list[np.ndarray]]:
    """Calculate facets of Wigner-Seitz cell. See plot_wigner_seitz_cell to plot the return values of this function

    Args:
        a1 (np.ndarray): First real-space lattice vector
        a2 (np.ndarray): Second real-space lattice vector
        a3 (np.ndarray): Third real-space lattice vector

    Returns:
        list[list[np.ndarray]]: List of facets. Each facet is a list of coordinates. Each coordinate is a numpy array.
                                Each facets is a face of the Wigner-Seitz cell
    """
    list_k_points = []
    for ii, jj, kk in itertools.product([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]):
        list_k_points.append(ii * a1 + jj * a2 + kk * a3)

    tess = Voronoi(list_k_points)
    out = []
    for r in tess.ridge_dict:
        if r[0] == 13 or r[1] == 13:
            out.append([tess.vertices[i] for i in tess.ridge_dict[r]])

    return out


def plot_wigner_seitz_cell(facets):
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection

    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    ax.set_box_aspect([1, 1, 1])

    poly = Poly3DCollection(facets, facecolors="lightblue", edgecolors="k", alpha=0.5)
    ax.add_collection3d(poly)

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    plt.tight_layout()
    plt.show()


def wigner_seitz_in_radius(facets) -> float:
    def plane_distance_from_origin(facet):
        v0, v1, v2 = map(np.array, facet[:3])
        e1 = v1 - v0
        e2 = v2 - v0
        n = np.cross(e1, e2)
        norm = np.linalg.norm(n)
        if norm < 1e-8:
            return None  # degenerate facet
        nhat = n / norm
        return abs(np.dot(nhat, v0))

    # Compute distances
    dists = [plane_distance_from_origin(f) for f in facets]
    dists = [d for d in dists if d is not None]

    # The in-radius
    r_in = min(dists)

    return r_in


def is_point_inside_parallelepiped(point: np.ndarray, v1: np.ndarray, v2: np.ndarray, v3: np.ndarray, tol=1e-8) -> tuple[bool, np.ndarray]:
    """Check if point is in a parallelepiped spanned by three vectors

    Args:
        point (np.ndarray): Point
        v1 (np.ndarray): First spanning vector
        v2 (np.ndarray): Second spanning vector
        v3 (np.ndarray): Third spanning vector

    Returns:
        tuple[bool, np.ndarray]: Tuple of: Is point in parallelepiped? [\alpha, \beta a2, \gamma] of point = \alpha v1 + \beta v2 + \gamma v3
    """
    M = np.vstack([v1, v2, v3]).T
    lambdas = np.linalg.solve(M, point)
    return np.all(lambdas >= 0 - tol) and np.all(lambdas <= 1 + tol), lambdas



import os
import sys
import time
import logging
import warnings
import numpy as np
import dopyqo
from dopyqo.colors import *


def snapshot_test(base_folder: str, prefix: str, active_electrons: int, active_orbitals: int, create: bool = False):
    snapshot_file = os.path.join("snapshots", f"eri_{prefix}_{active_electrons}e_{active_orbitals}o.npy")
    if not create:
        if not os.path.isfile(snapshot_file):
            print(f"{RED}Snapshot does not exist: {snapshot_file}{RESET_COLOR}")
            sys.exit(1)
        h_pqrs_ref = np.load(snapshot_file)

    dat_file = os.path.join(base_folder, f"{prefix}.save", "wfc1.dat")
    xml_file = os.path.join(base_folder, f"{prefix}.save", "data-file-schema.xml")
    pp_files = [
        os.path.join(base_folder, f"{prefix}.save", filename)
        for filename in os.listdir(os.path.join(base_folder, f"{prefix}.save"))
        if filename.lower().endswith(".upf")
    ]
    pps = [dopyqo.Pseudopot(pp_file) for pp_file in pp_files]

    binary_occupations = True

    wfc1_ncpp = dopyqo.Wfc.from_file(dat_file, xml_file, pseudopots=pps)

    overlaps_ncpp = wfc1_ncpp.get_overlaps()
    if not np.allclose(overlaps_ncpp, np.eye(overlaps_ncpp.shape[0])):
        warnings.warn("The wavefunctions are not orthonormal!")

    p = wfc1_ncpp.k_plus_G  # shape (#waves, 3)
    c_ip = wfc1_ncpp.evc  # shape (#bands, #waves)

    orbital_indices_core, orbital_indices_active = wfc1_ncpp.active_space(active_electrons=active_electrons, active_orbitals=active_orbitals)

    occs_tmp = wfc1_ncpp.occupations_binary if binary_occupations else wfc1_ncpp.occupations
    occ_str = " ".join(
        [
            (f"|{val}" if i == orbital_indices_active[0] else f"{val}|" if i == orbital_indices_active[-1] else f"{val}")
            for i, val in enumerate(occs_tmp)
        ]
    )
    print(f"The following active space will be considered:\n{occ_str}\tNumber of plane waves: {p.shape[0]}")

    occupations_full, c_ip_full = wfc1_ncpp.get_orbitals_by_index(
        orbital_indices_core + orbital_indices_active, binary_occupations=binary_occupations
    )
    occupations_active, c_ip_active = wfc1_ncpp.get_orbitals_by_index(orbital_indices_active, binary_occupations=binary_occupations)
    with warnings.catch_warnings():  # Ignore warning that all core orbitals are occupied
        warnings.filterwarnings("ignore", category=Warning)
        occupations_core, c_ip_core = wfc1_ncpp.get_orbitals_by_index(orbital_indices_core, binary_occupations=binary_occupations)

    # Calculate ERIs via pair density
    fft_grid = wfc1_ncpp.fft_grid
    # fft_grid = np.array([14, 14, 14])
    start_time = time.perf_counter()
    h_pqrs: np.ndarray = dopyqo.eri(c_ip=c_ip_active, b=wfc1_ncpp.b, mill=wfc1_ncpp.mill, fft_grid=fft_grid) / wfc1_ncpp.cell_volume
    time_eri = time.perf_counter() - start_time

    if not create:
        if not np.allclose(h_pqrs, h_pqrs_ref):
            print(f"{RED}Snapshot mismatch: {prefix}{RESET_COLOR}")
        else:
            print(f"{GREEN}Snapshot passed: {prefix}{RESET_COLOR}")
    else:
        np.save(snapshot_file, h_pqrs)
        print(f"{ORANGE}Snapshot created: {prefix}{RESET_COLOR}")


if __name__ == "__main__":
    create = False
    snapshot_test(base_folder=os.path.join("qe_files", "LiH"), prefix="LiH", active_electrons=4, active_orbitals=6, create=create)
    snapshot_test(base_folder=os.path.join("qe_files", "Mg"), prefix="Mg", active_electrons=20, active_orbitals=15, create=create)



import os
import dopyqo

input_file = os.path.join("qe_files", "LiH.scf.in")
cwd_before = os.getcwd()
print("Running QE...")
dopyqo.runQE(input_file=input_file, num_cpus=1, nk=1, nb=1, nt=1)
cwd_after = os.getcwd()
assert cwd_before == cwd_after
print("Test passed!")



from warnings import warn
import itertools
import logging
from scipy.special import erfc
import numpy as np
from numba import jit
from dopyqo.colors import *


def nuclear_repulsion_energy_ewald(
    atom_positions: np.ndarray,
    atomic_numbers: np.ndarray,
    lattice_vectors: np.ndarray,
    lattice_vectors_reciprocal: np.ndarray,
    cell_volume: float,
    gcut: float,
    sigma: float | None = None,
    rust_impl: bool = True,
) -> np.ndarray:
    r"""Calculate the nuclear repulsion energy E between all nuclei in a periodic lattice, i.e.,
    E = 1/2 \sum_I \sum_J \sum_T' Z_I Z_J / |R_I-R_J-T|
    where ' denotes that the term T=0 and I=J is omitted. Z_I are the charges of the nuclei
    and R_I their positions. T are all possible lattice translation vectors.
    This is equivalent to
    E = 1/2 \sum_I Z_I \Phi_I(R_I)
    where \Phi_I(r) is the electrostatic potential generated by all periodic nuclei EXCEPT
    the nucleus Z_I at position R_I (its periodic images are still included).
    To calculate \Phi_I(r) we define charge densities and solve the Poisson equation in
    real- and reciprocal-space. The nuclei Z_I are modelled by delta distribution at positions R_I+T.
    Additionally, Ewalds method is used, i.e., we add and substract Gaussian charge distributions
    at all R_I+T, and subtract a homogenous background charge simulating the electron-charges.
    A short-ranged term including all delta distributions minus all Gaussian charges (except the
    term where T=0 and I=J) can be calculated in real-space. A long-ranged term including
    all Gaussian charges at R_I+T minus the constant background charge can be calculated
    in reciprocal-space. The term describing the subtraction of the Gaussian charge at R_I
    is calculated in real-space and is the so-called self-energy term. The G -> 0 limit
    in the reciprocal space calculation is also calculated separately and is called the
    charged-system term. The value \sigma determining the splitting between the real-space
    and reciprocal-space term is calculated as done in Quantum ESPRESSO (see PW/src/ewald.f90:
    https://github.com/QEF/q-e/blob/de3035747f5d8f2ec9a67869827341ebb43f5b12/PW/src/ewald.f90)

    We calculate E = E_S + E_L + E_self + E_charged with

    Short-ranged term calculated in real-space
    E_S    = 1/2 \sum_T' \sum_I \sum_J Z_I Z_J / |R_I - R_J - T| erfc(|R_I - R_J - T| \sqrt(\sigma))
    where ' denotes that the term T=0 and I=J is omitted.

    Long-ranged term calculated in reciprocal-space
    E_L    = 1/2 4\pi/V \sum_{k \neq 0} \sum_I \sum_J Z_I Z_J / |k|^2 e^{i k . (R_I - R_J)} e^{- 1/(4 \sigma) |k|^2}

    Self-energy originated in subtracting the Gaussian charge at position R_I from the periodic charge density
    generated by all atoms (in real-space)
    E_self = - \sum_I (Z_I^2) \sqrt(\sigma) / \sqrt(\pi)

    Charged-system term from the G=0 term in the reciprocal-space sum
    E_charged = - \pi / (2 V \sigma)  (\sum_I Z_I)^2

    Notes:
    - When pseudopotentials are used Z_I are the valence charges, i.e. the charge of the nucleus that is not
      modelled with the pseudopotential.
    - This function should calculate the same Ewald energy calculated in Quantum ESPRESSO, called
      "ewald contribution" in the pw.x output file and "ewald" in the total_energy section in
      the output-xml.
    - Equivalent to Ewald calculation in PySCF of a cell in pyscf.pbc.get.cell.ewald function

    References:
    - http://metal.elte.hu/~groma/Anyagtudomany/kittel.pdf Appendix B
    - https://journals.aps.org/prb/pdf/10.1103/PhysRevB.53.1814, equation (16)
    - https://courses.physics.illinois.edu/phys466/sp2013/projects/2003/Team2/ewald_text.htm,
    - S. W. de Leeuw et al. Proc. R. Soc. Lond. A 1980 373, 27-56, doi: 10.1098/rspa.1980.0135,  equation (1.5)
    - M. P. Allen and D. J. Tildesley "Computer Simulation of Solids", equation (5.20)
    - http://micro.stanford.edu/mediawiki/images/4/46/Ewald_notes.pdf
    - https://homepages.uc.edu/~becktl/tlb_ewald.pdf
    - Quantum ESPRESSO source code in q-e/PW/src/ewald.f90

    Args:
        atom_positions (np.ndarray): Array of the coordinates R_I of every atom described
                                     by the pseudopotential. Shape (#atoms, 3)
        atomic_numbers (np.ndarray):  1D array of atomic number of each atom. Shape (#atoms,)
        lattice_vectors (np.ndarray): 2D array of lattice vectors. Each row is one lattice vector. Shape (3, 3)
        lattice_vectors_reciprocal (np.ndarray): 2D array of reciprocal lattice vectors. Each row is one reciprocal lattice vector. Shape (3, 3)
        cell_volume (float): Cell volume of the computationen real space cell.
        gcut (float): Cutoff of the reciprocal space sum in E_L, i.e., k < gcutrho. In Quantum ESPRESSO the value of the G-cutoff of the density is used.
        sigma (float | None, optional): Value for \sigma used in the Ewald summation to split real- and reciprocal-space sums. If None this is calculated to satisfy
                                        \sum_I 2 Z_I \sqrt{\frac\sigma\pi} \mathrm{erfc}\left(\sqrt{\frac{\left(4G_\mathrm{cut}\right)^2}{4\sigma}}\right)  \leq 10^{-7}
                                        Defaults to None.

    Raises:
        RuntimeError: If \sigma could not be calculated.

    Returns:
        float: Ewald energy
    """

    if rust_impl:
        try:
            import dopyqo_rs as calc_rs
        except ImportError:
            print(f"{ORANGE}Ewald warning: Could not import dopyqo_rs package. Falling back to python implementation.{RESET_COLOR}")
            rust_impl = False  # Set such that variable accurately represents if rust implementation is used
        else:  # No exception
            logging.info("Using Rust implementation.")
            return calc_rs.ewald(
                atom_positions,
                atomic_numbers,
                lattice_vectors,
                lattice_vectors_reciprocal,
                cell_volume,
                gcut,
            )

    if sigma is None:
        sigma = 2.8
        charge = np.sum(atomic_numbers)
        gcutm = gcut**2
        #
        # choose sigma in order to have convergence in the sum over G
        # upperbound is a safe upper bound for the error in the sum over G
        #
        while True:
            if sigma <= 0.0:
                raise RuntimeError("Optimal sigma for Ewald sum not found!")
            upperbound = 2.0 * charge**2 * np.sqrt(sigma / np.pi) * erfc(np.sqrt(gcutm / 4.0 / sigma))
            if upperbound > 1e-7:
                sigma = sigma - 0.1
            else:
                break

    logging.info("sigma %f", sigma)

    e_short = 0.0  # real-space sum
    converged_real = False
    # n_max = 5
    n_max_used = 0

    alat = np.linalg.norm(lattice_vectors[0], ord=2)
    t_vec_max_norm = 4.0 / np.sqrt(sigma) / alat
    b_norms = np.linalg.norm(lattice_vectors_reciprocal, ord=2, axis=1)
    n_max_x, n_max_y, n_max_z = (b_norms * t_vec_max_norm).astype(int) + 2

    # Generate list or translation vectors ordered by their norm
    # n_unordered = itertools.product(range(-n_max, n_max + 1), repeat=3)
    logging.debug("Generating unordered translation vectors...")
    t_vecs_unordered = [
        [np.dot(lattice_vectors, np.array([nx, ny, nz])), [nx, ny, nz]]
        for nx, ny, nz in itertools.product(
            range(-n_max_x, n_max_x + 1),
            range(-n_max_y, n_max_y + 1),
            range(-n_max_z, n_max_z + 1),
        )
        # for nx, ny, nz in n_unordered
    ]

    t_vecs = sorted(t_vecs_unordered, key=lambda x: np.linalg.norm(x[0], ord=2))

    for t_vec, (nx, ny, nz) in t_vecs:
        for i, pos_i in enumerate(atom_positions):  # \sum_I
            z_i = atomic_numbers[i]
            for j, pos_j in enumerate(atom_positions):  # \sum_J
                if np.all(t_vec == 0.0) and i == j:
                    continue
                z_j = atomic_numbers[j]
                r = np.linalg.norm(pos_i - pos_j - t_vec, ord=2)
                e_short += z_i * z_j * erfc(r * np.sqrt(sigma)) / r

                # with this choice terms up to ZiZj*erfc(4) are counted (erfc(4)=1.5e-8)
                if r >= 4 / np.sqrt(sigma):
                    # print(f"Stopping real-sum at {nx}/{ny}/{nz} (sigma={sigma})")
                    converged_real = True
                    n_max_used = np.max([np.abs(nx), np.abs(ny), np.abs(nz), n_max_used])
                    # break
        # if converged_real:
        #     break
    if converged_real:
        logging.info("Real-space sum converged! Largest n was %i", n_max_used)
    else:
        logging.warning("Real-space sum NOT converged!")

    e_short *= 1 / 2

    e_long = 0.0  # reciprocal-space sum

    # Estimate size of Miller indices grid using the cutoff-energy and reciprocal lattice vectors
    # A simple, but not accurate enough, estimate would be gcutrho/|b_i| for i \in {1, 2, 3}
    # Here we take the shape of the reciprocal lattice into account:
    #       max[ gcutrho/(|b_i| + \sum_{j \neq i} b_i . b_j ) ] for i \in {1, 2, 3}
    # TODO: To use the same grid QE uses we have to read the Miller indices used for the density,
    #       which is probably saved in charge-density.dat.
    #       Alternatively check how pymatgen chooses the reciprocal grid in its Ewald calculation method
    b1 = lattice_vectors_reciprocal[0]
    b2 = lattice_vectors_reciprocal[1]
    b3 = lattice_vectors_reciprocal[2]
    mill_max = max(
        [
            round(abs(gcut / (np.linalg.norm(b1) + np.dot(b1, b2) + np.dot(b1, b3)))),
            round(abs(gcut / (np.linalg.norm(b2) + np.dot(b2, b1) + np.dot(b2, b3)))),
            round(abs(gcut / (np.linalg.norm(b3) + np.dot(b3, b1) + np.dot(b3, b2)))),
        ]
    )
    x, y, z = np.meshgrid(
        np.arange(-mill_max, mill_max + 1),
        np.arange(-mill_max, mill_max + 1),
        np.arange(-mill_max, mill_max + 1),
        indexing="ij",
    )
    mill_rho = np.stack((x.ravel(), y.ravel(), z.ravel()), axis=1)
    k_vecs = np.einsum("ij, kj -> ki", lattice_vectors_reciprocal.T, mill_rho)
    norms = np.linalg.norm(k_vecs, axis=1)
    k_vecs = k_vecs[norms <= gcut]
    for k_vec in k_vecs:
        k2 = np.linalg.norm(k_vec, ord=2) ** 2
        if np.isclose(k2, 0.0):
            continue
        structure_factor = np.sum(atomic_numbers * np.exp(1.0j * np.dot(atom_positions, k_vec)))
        e_long += np.exp(-k2 / (4 * sigma)) * np.abs(structure_factor) ** 2 / k2

    e_long *= 4.0 * np.pi / 2.0 / cell_volume

    logging.debug("Reciprocal-space sum calculated!")

    # Term arising from substracting the Gaussian charge distribution at R_I
    # in real-space
    e_self = -np.sum(atomic_numbers**2) * np.sqrt(sigma / np.pi)

    # The analytic limit for G → 0 in the reciprocal sum
    # The G=0 term in the reciprocal term is only partly cancelled by the uniform
    # background charge. The remaining term is the following
    # Also see Martin, 2020, "Electronic Structure" eq. (F.6) at the text afterwards
    # which unfortunately misses a derivation
    e_charged = -np.pi / cell_volume / sigma * np.sum(atomic_numbers) ** 2 / 2
    # e_charged = -np.sum(atomic_numbers) ** 2 / sigma / 2 * np.pi / cell_volume

    e_tot = e_short + e_long + e_self + e_charged

    logging.info(
        "e_short: %f, e_long: %f, e_self: %f, e_charged: %f, e_tot: %f",
        e_short,
        e_long,
        e_self,
        e_charged,
        e_tot,
    )

    return e_tot


def check_symmetry_one_body_matrix(matrix: np.ndarray):
    """Check if given matrix satisfies the symmetries of one-body matrices (hermitian)

    Args:
        matrix (np.ndarray): Matrix to check symmetries of

    Returns:
        bool: Bool specifying of symmetry is satisfied
    """
    matrix_hermitian = matrix.T.conj()
    allclose_hermitian = np.allclose(matrix, matrix_hermitian)
    return allclose_hermitian


def check_symmetry_two_body_matrix(matrix: np.ndarray):
    """Check if given matrix satisfies the symmetries of ERIs

    Args:
        matrix (np.ndarray): Matrix to check symmetries of

    Returns:
        tuple[bool, bool, bool]: Bools specifying if following symmetries are satisfied: swap symmetry, hermitian symmetry, hermitian+swap symmetry
    """
    matrix_swap = matrix.swapaxes(0, 1).swapaxes(2, 3)
    matrix_hermitian = matrix.T.conj()
    matrix_hemitian_swap = matrix.T.conj().swapaxes(0, 1).swapaxes(2, 3)

    allclose_swap = np.allclose(matrix, matrix_swap)  # , rtol=1e-5, atol=1e-5)
    allclose_hermitian = np.allclose(matrix, matrix_hermitian)  # , rtol=1e-5, atol=1e-5)
    allclose_hermitian_swap = np.allclose(matrix, matrix_hemitian_swap)  # , rtol=1e-5, atol=1e-5)

    allclose_ljki = None
    allclose_ikjl = None
    allclose_kilj = None
    allclose_jlik = None
    if np.allclose(matrix.real, matrix):
        logging.info("Two-body matrix is real. Checking real-symmetries...")
        allclose_hermitian = np.allclose(matrix, matrix.T)  # ijkl = lkji
        allclose_swap = np.allclose(matrix, matrix.transpose(1, 0, 3, 2))  # ijkl = jilk
        allclose_ljki = np.allclose(matrix, matrix.transpose(3, 1, 2, 0))  # ijkl = ljki
        allclose_ikjl = np.allclose(matrix, matrix.transpose(0, 2, 1, 3))  # ijkl = ikjl
        allclose_hermitian_swap = np.allclose(matrix, matrix.T.transpose(1, 0, 3, 2))  # ijkl = klij
        allclose_kilj = np.allclose(matrix, matrix.transpose(2, 0, 3, 1))  # ijkl = kilj
        allclose_jlik = np.allclose(matrix, matrix.transpose(1, 3, 0, 2))  # ijkl = jlik

    return (
        allclose_swap,
        allclose_hermitian,
        allclose_hermitian_swap,
        allclose_ljki,
        allclose_ikjl,
        allclose_kilj,
        allclose_jlik,
    )


def iTj(p: np.ndarray, c_ip: np.ndarray) -> np.ndarray:  # Calculates <i|T|j>
    """Calculate kinetic energy matrix in Hartree units in the Kohn-Sham basis

    Args:
        p (np.ndarray): Array of momentum vectors, shape (#waves, 3)
        c_ip (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals
            in the plane wave basis, shape (#bands, #waves)

    Returns:
        np.ndarray: Kinetic energy matrix
    """
    # Kinetic energy matrix in Hartree units
    p_norm = p[:, 0] ** 2 + p[:, 1] ** 2 + p[:, 2] ** 2  # p^2
    # 1/2 <i|p^2|j>
    return 0.5 * np.einsum("ip, p, jp -> ij", c_ip.conjugate(), p_norm, c_ip)


def iUj(
    p: np.ndarray,
    c_ip: np.ndarray,
    atom_positions: np.ndarray,
    atomic_numbers: np.ndarray,
    cell_volume: float,
) -> np.ndarray:
    r"""Calculate nuclear interaction (U(r) = \sum_i Z_i / |R_I-r|) matrix in Hartree units in the Kohn-Sham basis

    Args:
        p (np.ndarray): Array of momentum vectors
        c_ip (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals in the plane wave basis
        atom_positions (np.ndarray): 2D array of positions of each atom. Shape (#atoms, 3)
        atomic_numbers (np.ndarray): 1D array of atomic number of each ato. Shape (#atoms,)m
        cell_volume (float): Cell volume of the computationen real space cell.

    Returns:
        np.ndarray: Nuclear interaction matrix
    """
    # Nuclear interaction matrix in Hartree units
    # Calculates <i|U|j>
    # <p|U|q> = 4\pi / cell_volume \sum_I Z_I  exp(-i (p-q) . R_I) 1/(q-p)²
    # <i|U|j> = \sum_{p,q,p!=q}  c_{p,i}^* c_{q,j} U_pq = 4\pi / cell_volume \sum_{p,q,p!=q} \sum_I Z_I  c_{p,i}^* c_{q,j} exp(-i (p-q) . R_I) 1/(q-p)²
    q = p
    Z_I = atomic_numbers  # shape (#atoms,)
    R_I = atom_positions  # shape (#atoms, 3)

    q = p
    p_minus_q = p[:, None] - q[None]  # shape (#waves, #waves, 3)
    # q[None] is of shape (1, #waves, 3) while p[:, None] is of shape (#waves, 1, 3)
    # so p_minus_q[i,j] = p[i] - q[j]

    p_minus_q = q[:, None] - p[None]  # shape (#waves, #waves, 3)
    p_minus_q_norm = np.linalg.norm(p_minus_q, ord=2, axis=2)  # Has zeros on diagonal, shape (#waves, #waves)
    p_minus_q_norm_filtered = p_minus_q_norm.copy() + np.eye(p_minus_q_norm.shape[0])  # Replace zeros on diagonal with ones

    p_minus_q_dot_R = np.sum(p_minus_q[None] * R_I[:, None, None], axis=3)  # sum over 3D-coordinates, shape (#atoms, #waves, #waves)
    # \sum_I 4\pi/V e^{-i (G-G').R_I}
    prefactor = 4 * np.pi / cell_volume * np.sum(Z_I[:, None, None] * np.exp(-1j * p_minus_q_dot_R), axis=0)  # sum over atoms, shape (#waves, #waves)
    one_over_p_minus_q_norm_squared = 1 / p_minus_q_norm_filtered**2
    one_over_p_minus_q_norm_squared = one_over_p_minus_q_norm_squared.copy() - np.eye(one_over_p_minus_q_norm_squared.shape[0])

    u_mat = prefactor * one_over_p_minus_q_norm_squared
    assert (u_mat.diagonal() == 0.0).all()

    return c_ip.conj() @ u_mat @ c_ip.T



import numpy as np
from qiskit.quantum_info import Statevector
from pyscf.fci import FCIvector


def largest_binary_with_same_number_of_ones(binary_string: str) -> bytes:
    count_ones = binary_string.count("1")
    return "1" * count_ones + "0" * (len(binary_string) - count_ones)


def closest_larger_binary_with_same_number_of_ones(binary_string: str) -> bytes | None:
    # Count the number of 1's in the given binary string
    count_ones = binary_string.count("1")

    # Find the next larger integer with the same number of 1's
    next_integer = int(binary_string, 2) + 1
    while bin(next_integer).count("1") != count_ones:
        next_integer += 1

    if next_integer > int(largest_binary_with_same_number_of_ones(binary_string), base=2):
        return None

    new_binary_string = bin(next_integer)[2:]

    if len(new_binary_string) < len(binary_string):
        n_missing_zeros = len(binary_string) - len(new_binary_string)
        new_binary_string = f"{'0'*n_missing_zeros}{new_binary_string}"

    assert (l1 := len(new_binary_string)) == (l2 := len(binary_string)), (
        f"New binary string {new_binary_string} (length: {l1}) has not the same length "
        + f"as the original binary string binary_string {binary_string} (length: {l2}). "
        + "This should not be possible! Please, contact a developer!"
    )

    # Convert the next integer to binary and return it
    return new_binary_string


def has_same_spin(binary_string: str, binary_string_reference: str) -> bool:
    count_ones = binary_string_reference.count("1")
    ones_correct = binary_string.count("1") == count_ones
    if not ones_correct:
        return False

    count_spin_up = binary_string_reference[::2].count("1")
    count_spin_dw = binary_string_reference[1::2].count("1")
    spin_up_correct = binary_string[::2].count("1") == count_spin_up
    spin_dw_correct = binary_string[1::2].count("1") == count_spin_dw

    return ones_correct and spin_up_correct and spin_dw_correct


def closest_larger_binary_with_same_spin(binary_string: str) -> bytes | None:
    # Find the next larger integer with the same spin
    next_integer = int(binary_string, 2) + 1
    while not has_same_spin(bin(next_integer)[2:], binary_string) and next_integer < int(
        largest_binary_with_same_number_of_ones(binary_string), base=2
    ):
        next_integer += 1

    if next_integer > int(largest_binary_with_same_number_of_ones(binary_string), base=2):
        return None

    new_binary_string = bin(next_integer)[2:]

    if len(new_binary_string) < len(binary_string):
        n_missing_zeros = len(binary_string) - len(new_binary_string)
        new_binary_string = f"{'0'*n_missing_zeros}{new_binary_string}"

    assert (l1 := len(new_binary_string)) == (l2 := len(binary_string)), (
        f"New binary string {new_binary_string} (length: {l1}) has not the same length "
        + f"as the original binary string binary_string {binary_string} (length: {l2}). "
        + "This should not be possible! Please, contact a developer!"
    )

    # Convert the next integer to binary and return it
    return new_binary_string


def statevector_from_fci(fci_vector: FCIvector, nelec: tuple[int, int], norb: int) -> Statevector:
    base_str_up = "0" * (norb - nelec[0]) + "1" * nelec[0]
    base_str_dw = "0" * (norb - nelec[1]) + "1" * nelec[1]

    # to int
    base_int_up = int(base_str_up, base=2)
    base_int_dw = int(base_str_dw, base=2)

    # to binary with length of norb
    # bin(base_int_up)[2:].zfill(norb)

    strs_up = [base_str_up]
    while (val := closest_larger_binary_with_same_number_of_ones(strs_up[-1])) is not None:
        # print(f"Appending {val}")
        strs_up.append(val)

    strs_dw = [base_str_dw]
    while (val := closest_larger_binary_with_same_number_of_ones(strs_dw[-1])) is not None:
        # print(f"Appending {val}")
        strs_dw.append(val)

    state_vec = np.zeros((2 ** (norb * 2),))
    for i, str_up in enumerate(strs_up):
        for j, str_dw in enumerate(strs_dw):
            idx = int(str_up + str_dw, base=2)
            state_vec[idx] = fci_vector[i, j]

    return Statevector(state_vec, dims=state_vec.shape[0])

    # strs_up = [base_str_up]
    # strs_up = strs_up + [val for val in iter(lambda: closest_larger_binary_with_same_number_of_ones(strs_up[-1]), None)]


def statevector_from_civector_restricted(ffsim_state: np.ndarray, nelec: tuple[int, int], norb: int) -> np.ndarray:
    base_str = "0" * (norb * 2 - np.sum(nelec)) + "1" * np.sum(nelec)
    # print(f"base_str: {base_str}")

    strs = [base_str]
    while (val := closest_larger_binary_with_same_spin(strs[-1])) is not None:
        # print(f"Appending {val}")
        strs.append(val)

    if len(strs) < len(ffsim_state):
        raise ValueError(
            f"The provided ffsim state describes more states than expected. {len(strs)} spin-restricted states found but ffsim state has {len(ffsim_state)}. Maybe you wanted to use statevector_from_ffsim_state_general?"
        )
    elif len(strs) > len(ffsim_state):
        raise ValueError(
            f"The provided ffsim state describes less states than expected. {len(strs)} spin-restricted states found but ffsim state has {len(ffsim_state)}."
        )

    state_vec = np.zeros((2 ** (norb * 2),))
    for i, str in enumerate(strs):
        idx = int(str, base=2)
        state_vec[idx] = ffsim_state[i]

    return Statevector(state_vec, dims=state_vec.shape[0])


def statevector_from_civector_general(ffsim_state: np.ndarray, nelec: tuple[int, int], norb: int) -> np.ndarray:
    base_str = "0" * (norb * 2 - np.sum(nelec)) + "1" * np.sum(nelec)
    # print(f"base_str: {base_str}")

    strs = [base_str]
    while (val := closest_larger_binary_with_same_number_of_ones(strs[-1])) is not None:
        # print(f"Appending {val}")
        strs.append(val)

    if len(strs) < len(ffsim_state):
        raise ValueError(
            f"The provided ffsim state describes more states than expected. {len(strs)} spin-restricted states found but ffsim state has {len(ffsim_state)}."
        )
    elif len(strs) > len(ffsim_state):
        raise ValueError(
            f"The provided ffsim state describes less states than expected. {len(strs)} spin-restricted states found but ffsim state has {len(ffsim_state)}. Maybe you wanted to use statevector_from_ffsim_state_restricted?"
        )

    state_vec = np.zeros((2 ** (norb * 2),))
    for i, str in enumerate(strs):
        idx = int(str, base=2)
        state_vec[idx] = ffsim_state[i]

    return Statevector(state_vec, dims=state_vec.shape[0])



import numpy as np
import pyvista as pv
from dopyqo.colors import *
from dopyqo.xsf import load_xsf


def plot_3d_data(
    coordinates: np.ndarray,
    data: np.ndarray,
    isosurfaces: int | list[float],
    isovals_as_percent: bool = False,
    atom_positions: np.ndarray | None = None,
    atom_labels: list[str] | None = None,
    extend_data: bool = False,
    grid_vectors: np.ndarray | None = None,
    cmap: str = "coolwarm",
    atom_color: str = "#0dff00",
    atom_radius: float = 0.3,
    opacity: float = 1.0,
    filtered_regions: int = 0,
    point_of_interest: np.ndarray = np.array([0.0, 0.0, 0.0]),
    html_filename: str = "3d_pyvista.html",
    plotter: pv.Plotter | None = None,
    meshes: list[pv.PolyData | pv.DataSet | pv.MultiBlock | np.ndarray] = [],
):
    """Plot 3D data given by coordinates and scalar values for each coordinate (data).
    The coordinates need to be structured in a way that we can create a PyVista/VTK StructuredGrid from them.
    The data does not need to be defined on a orthogonal grid. See the definition/usage of PyVista/VTK StructuredGrid for more information.
    At the end the plot will be saved to a html-file.

    Args:
        coordinates (np.ndarray): Numpy array of coordinates of shape (nx,ny,nz,3) or of shape (nx*ny*nz,3)
        data (np.ndarray): Numpy array of scalar field data of shape (nx,ny,nz) or of shape (nx*ny*nz)
        isosurfaces (int | list[float]): isosurfaces argument passed to PyVista mesh.contour function
        isovals_as_percent (bool): Should the isosurface values (isosurfaces argument) be interpreted as percentage value relative to
                                   the minimum and maximum of the data values. Can only be set to true if isosurfaces argument is a list.
        atom_positions (np.ndarray | None, optional): Numpy array of atom positions of shape (N,3). If None no atoms will be plotted.
                                                      Defaults to None.
        extend_data (bool): Whether to extend the data along each grid vector once. If True grid_vectors has be given as numpy array. Defaults to False.
        grid_vectors (np.ndarray | None): Numpy array of vectors defining the grid, e.g. lattice vectors, of shape (3,3), where each row is one vector.
                                          Has not to be numpy array if extend_data is True. Defaults to None.
        cmap (str, optional): cmap argument passed to PyVista Plotter.add_mesh function for plotting the isosurfaces. Defaults to "coolwarm".
        atom_color (str, optional): color argument passed to PyVista Plotter.add_mesh for plotting the atom spheres. Defaults to "#0dff00".
        opacity (float, optional): Opacity value of the shown isosurfaces. Defaults to 1.0.
        filtered_regions (int, optional): Number of isosurfaces around point_of_interest that are highlighted while all others are plotted with opacity*0.2.
                                          Can be slow. If set to 0 all isosurfaces are plotted. Defaults to 0.
        point_of_interest (np.ndarray, optional): point_of_interest used for filtering isosurfaces. See filtered_regions argument. Defaults to np.array([0.0, 0.0, 0.0]).
        html_filename (str, optional): Filename of the html-file the plot will be saved to. Postfix ".html" is appended if not already present. Defaults to "3d_pyvista.html".
        plotter (pv.Plotter | None, optional): PyVista Plotter object used for plotting. If None a new object will be created. Defaults to None.
        meshes (list[pv.PolyData | pv.DataSet | pv.MultiBlock | np.ndarray]): List of additional meshes that are plotted. Are each passed to plotter.add_mesh function.
                                                                              Defaults to [].
    """
    if not html_filename.endswith(".html"):
        html_filename += ".html"
    if atom_labels is None and atom_positions is not None:
        atom_labels = ["" for _ in atom_positions]
    if isovals_as_percent and not isinstance(isosurfaces, list):
        print(
            f"{ORANGE}Plotting warning: Argument isovals_as_percent is set to true but argument isosurfaces is not a list. Setting isovals_as_percent to False!{RESET_COLOR}"
        )
        isovals_as_percent = False

    if isovals_as_percent:
        isosurfaces = [data.min() + x * data.ptp() for x in isosurfaces]

    nx, ny, nz = data.shape

    mesh = pv.StructuredGrid()
    mesh.points = coordinates.reshape((-1, 3))
    # If dimensions are not specified the mesh remains "empty" and cannot be plotted
    mesh.dimensions = [nx, ny, nz]
    mesh.point_data["values"] = data.ravel()
    contours = mesh.contour(isosurfaces=isosurfaces)

    extended_contours = []
    if extend_data:
        if grid_vectors is None:
            print(
                f"{RED}Plot warning: extend_data is set to True but grid_vectors are not given. Cannot extend data without grid vectors.{RESET_COLOR}"
            )
        else:
            coordinates_list_tmp = []
            coordinates_tmp = coordinates.copy()
            coordinates_tmp += grid_vectors[None, 0]
            coordinates_list_tmp.append(coordinates_tmp)  # shifted in first grid direction
            coordinates_tmp = coordinates.copy()
            coordinates_tmp += grid_vectors[None, 1]
            coordinates_list_tmp.append(coordinates_tmp)  # shifted in second grid direction
            coordinates_tmp = coordinates.copy()
            coordinates_tmp += grid_vectors[None, 2]
            coordinates_list_tmp.append(coordinates_tmp)  # shifted in third grid direction
            coordinates_tmp = coordinates.copy()
            coordinates_tmp += grid_vectors[None, 0]
            coordinates_tmp += grid_vectors[None, 1]
            coordinates_list_tmp.append(coordinates_tmp)  # shifted in first + second grid direction
            coordinates_tmp = coordinates.copy()
            coordinates_tmp += grid_vectors[None, 0]
            coordinates_tmp += grid_vectors[None, 2]
            coordinates_list_tmp.append(coordinates_tmp)  # shifted in first + third grid direction
            coordinates_tmp = coordinates.copy()
            coordinates_tmp += grid_vectors[None, 1]
            coordinates_tmp += grid_vectors[None, 2]
            coordinates_list_tmp.append(coordinates_tmp)  # shifted in second + third grid direction
            coordinates_tmp = coordinates.copy()
            coordinates_tmp += grid_vectors[None, 0]
            coordinates_tmp += grid_vectors[None, 1]
            coordinates_tmp += grid_vectors[None, 2]
            coordinates_list_tmp.append(coordinates_tmp)  # shifted in first + second + third grid direction

            for coordinates_tmp in coordinates_list_tmp:
                mesh_tmp = pv.StructuredGrid()
                mesh_tmp.points = coordinates_tmp.reshape((-1, 3))
                mesh_tmp.dimensions = [nx, ny, nz]
                mesh_tmp.point_data["values"] = data.ravel()
                contours_tmp = mesh_tmp.contour(isosurfaces=isosurfaces)
                extended_contours.append(contours_tmp)

    ##############################################################
    #          Define PyVista spheres for atom positions         #
    ##############################################################
    # NOTE: Point cloud would also be possible but did not render as spheres on our machine
    spheres = []
    if atom_positions is not None:
        spheres = [pv.Sphere(radius=atom_radius, center=pos) for pos in atom_positions]

    ##############################################################
    #                     Filter isosurfaces                     #
    ##############################################################
    if filtered_regions != 0:
        if extend_data:
            print(
                f"{ORANGE}Plot warning: You specified to filter regions ({filtered_regions=}) and to extend the data ({extend_data=}). "
                + f"Beware that currently isosurfaces in the extend data are not filtered.{RESET_COLOR}"
            )
        all_regions = contours.connectivity("all")
        region_ids = np.unique(all_regions["RegionId"])
        my_region_ids = region_ids[:filtered_regions]
        my_regions = contours.connectivity("specified", my_region_ids)
        # NOTE: Calculating distances like in the following takes quite long. How to speed up?
        distances = [
            np.min(np.linalg.norm(contours.connectivity("specified", region_id).points - point_of_interest, axis=1)) for region_id in region_ids
        ]
        my_region_ids = np.argsort(distances)[:2]
        my_regions = contours.connectivity("specified", my_region_ids)

        largest_region = contours.connectivity("largest")

    ##############################################################
    #                          Plotting                          #
    ##############################################################
    # contours = mesh.contour(isosurfaces=7)
    if plotter is None:
        plotter = pv.Plotter(off_screen=True)
    plotter.add_mesh(mesh.outline())
    if filtered_regions != 0:
        plotter.add_mesh(contours, show_edges=False, cmap=cmap, opacity=opacity * 0.2)
        plotter.add_mesh(my_regions, show_edges=False, cmap=cmap, opacity=opacity)
    else:
        plotter.add_mesh(contours, show_edges=False, cmap=cmap, opacity=opacity, label="data")
    for contour_tmp in extended_contours:
        plotter.add_mesh(contour_tmp, show_edges=False, cmap=cmap, opacity=opacity)
    for sphere_tmp, label_tmp in zip(spheres, atom_labels):
        label_tmp = label_tmp.strip()
        plotter.add_mesh(sphere_tmp, color=atom_color)
        text_height = 0.2 * len(label_tmp.splitlines())
        text_depth = text_height / 5.0
        text_tmp = pv.Text3D(label_tmp, center=[0, 0, 0], height=text_height, depth=text_depth)
        width_tmp = text_tmp.bounds.x_max - text_tmp.bounds.x_min
        height_tmp = text_tmp.bounds.y_max - text_tmp.bounds.y_min
        label_pos = np.array(sphere_tmp.center) + np.array([atom_radius + width_tmp / 2, atom_radius + height_tmp / 2, 0.0])
        text_tmp = pv.Text3D(label_tmp, center=label_pos, height=text_height, depth=text_depth, normal=[0, 0, 1])
        plotter.add_mesh(text_tmp, color="black")
    for mesh_tmp in meshes:
        plotter.add_mesh(mesh_tmp)
    # text_tmp = pv.Text3D("Test text", center=(-10.0, 0.0, 10.0))
    # plotter.add_mesh(text_tmp)
    # # Label not showing in interactive/html view, see:
    # # https://github.com/pyvista/pyvista/issues/6202
    # label = pv.Label(text="Test text", position=(-10.0, 0.0, 10.0), size=50)
    # plotter.add_actor(label)
    plotter.show_grid()
    plotter.show_axes()
    plotter.export_html(html_filename)
    print(f"{GREEN}Plot sucessfully saved to {html_filename}.{RESET_COLOR}")

    # NOTE: Volume plot only works for numpy array, ImageData, RectilinearGrid or UnstructuredGrid
    # plotter.add_volume(xsf_data, scalars="values", cmap="magma", opacity=[1,0,1])
    # plotter.add_volume(mesh, scalars="orbital", cmap="magma", opacity=[1,0,1])

    # # Plot mesh itself without the scalar data associated to each grid point
    # # This shows the shape of the unit cell and each voxel in the dataset
    # plotter = pv.Plotter(off_screen=True)
    # plotter.add_mesh(mesh)
    # plotter.add_points(mesh.points, color="red", point_size=2)
    # plotter.add_mesh(
    #     mesh.get_cell(mesh.n_cells - 1).cast_to_unstructured_grid(),
    #     color="pink",
    #     edge_color="blue",
    #     line_width=5,
    #     show_edges=True,
    # )
    # plotter.show_grid()
    # plotter.show_axes()
    # plotter.export_html("contour_plot.html")

    return plotter


def plot_xsf(
    filename: str,
    isosurfaces: int | list[float],
    isovals_as_percent: bool = False,
    cmap: str = "coolwarm",
    color_atoms: str = "#0dff00",
    opacity: float = 1.0,
    filtered_regions: int = 0,
    point_of_interest: np.ndarray = np.array([0.0, 0.0, 0.0]),
    plot_lattice_vectors: bool = True,
    html_filename: str = "3d_pyvista.html",
    plotter: pv.Plotter | None = None,
    plot_abs: bool = False,
):
    xsf_data, xsf_grid_vecs, xsf_origin, xsf_atom_pos, xsf_primvecs, xsf_convvecs = load_xsf(filename=filename)
    if plot_abs:
        xsf_data = np.abs(xsf_data)
    # Each row of xsf_grid_vecs is one grid_vector

    ##############################################################
    #              Calculate data point coordinates              #
    ##############################################################
    nx, ny, nz = xsf_data.shape
    coordinates = (
        np.arange(nx)[:, None] * xsf_grid_vecs[0] / nx
        + np.arange(ny)[:, None, None] * xsf_grid_vecs[1] / ny
        + np.arange(nz)[:, None, None, None] * xsf_grid_vecs[2] / nz
        + xsf_origin
    )

    lattice_vectors_meshes = []
    if plot_lattice_vectors:
        lattice_vectors_meshes = [pv.Arrow(start=xsf_origin, direction=v, shaft_radius=2e-2, tip_radius=5e-2, scale="auto") for v in xsf_grid_vecs]

    from dopyqo.plotting import plot_3d_data

    return plot_3d_data(
        coordinates=coordinates,
        data=xsf_data,
        isosurfaces=isosurfaces,
        isovals_as_percent=isovals_as_percent,
        atom_positions=xsf_atom_pos,
        meshes=lattice_vectors_meshes,
        cmap=cmap,
        atom_color=color_atoms,
        opacity=opacity,
        filtered_regions=filtered_regions,
        point_of_interest=point_of_interest,
        html_filename=html_filename,
        plotter=plotter,
    )



import numpy as np


def transform_one_body_matrix(one_body_matrix: np.ndarray, transform_matrix: np.ndarray) -> np.ndarray:
    assert one_body_matrix.ndim == 2
    assert transform_matrix.ndim == 2
    assert one_body_matrix.shape == (one_body_matrix.shape[0],) * 2
    assert transform_matrix.shape == (transform_matrix.shape[0],) * 2

    # <p|V|q> = \sum_{i,j} <i|p>* <i|V|j> <j|q>
    # where <i|p> is transform_matrix
    # and <i|V|j> is one_body_matrix
    res = np.einsum("ip, ij, jq -> pq", transform_matrix.conj(), one_body_matrix, transform_matrix)
    return res


def transform_two_body_matrix(two_body_matrix: np.ndarray, transform_matrix: np.ndarray) -> np.ndarray:
    assert two_body_matrix.ndim == 4
    assert transform_matrix.ndim == 2
    assert two_body_matrix.shape == (two_body_matrix.shape[0],) * 4
    assert transform_matrix.shape == (transform_matrix.shape[0],) * 2

    # <pq|V|rs> = \sum_{i,j,k,l} <i|p>* <j|q>* <ij|V|kl> <k|r> <l|s>
    # where <i|p> is transform_matrix
    # and <ij|V|kl> is two_body_matrix
    res = np.einsum(
        "ip, jq, ijkl, kr, ls -> pqrs",
        transform_matrix.conj(),
        transform_matrix.conj(),
        two_body_matrix,
        transform_matrix,
        transform_matrix,
    )
    return res


def to_density_matrix(active_electrons, active_orbitals):
    diag_density = []
    electrons = active_electrons
    orbitals = active_orbitals
    while electrons > 0 and len(diag_density) < orbitals:
        diag_density.append(2)
        electrons -= 2

    diag_density.extend([0] * (orbitals - len(diag_density)))
    density_mat = np.diag(diag_density)

    return density_mat



"""
Dopyqo: Many-body analysis on top of Quantum ESPRESSO calculations
"""

from dopyqo.info import __version__, HOMEPAGE
from dopyqo.helpers.vqe_helpers import *
from dopyqo.helpers.printing import *
from dopyqo.helpers.config import *
from dopyqo.helpers.matrix_elements import *
from dopyqo.helpers.atoms import elements_to_atomic_number
from dopyqo.helpers.tcc_helpers import *
from dopyqo.calc_matrix_elements import nuclear_repulsion_energy_ewald, iTj
from dopyqo.eri_pair_densities import (
    eri,
    get_frozen_core_energy_pp,
    get_frozen_core_energy_given_pp,
    get_frozen_core_pot_and_energy_given_pp,
    get_frozen_core_pot,
)
from dopyqo.calc_pseudo_pot import calc_pps
from dopyqo.pseudopot import Pseudopot
from dopyqo.units import *
from dopyqo.wfc import Wfc
from dopyqo.hamiltonian import Hamiltonian
from dopyqo.wannier90 import read_u_mat
from dopyqo.transform_matrices import transform_one_body_matrix, transform_two_body_matrix, to_density_matrix
from dopyqo.wfc import runQE
from dopyqo.scripts.main import run



import os
import numpy as np
import dopyqo
import dopyqo.wannier90

if __name__ == "__main__":
    base_folder = os.path.join("qe_files", "Be_wannier")
    filename_umat = os.path.join(base_folder, "Be_u.mat")

    kpt_to_u = dopyqo.read_u_mat(filename=filename_umat)

    filename_hr = os.path.join(base_folder, "Be_hr.dat")

    hr_mat = dopyqo.wannier90.read_hr_dat(filename=filename_hr)
    # KS eigenvalues and eigenvectors in the Wannier basis
    eigvals, eigvecs = np.linalg.eig(hr_mat[:, :, 0])

    # Check unitarity
    u_mat = np.stack(list(kpt_to_u.values()), axis=2)
    print(f"Unitary: {dopyqo.wannier90.check_unitarity_u(u_mat[:, :, 0])}")
